/*
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA
*/

/*
  Maintainance log started on April 30th, 2002

  April 30th, 2002   - added support for time sequence attributes,
                     - added debug info
                     - cleaned up formatting -kap

  June 12th, 2002    - added checkSubsets method to generate subsets
		       of size k-1 of an itemset of size k and check 
		       that they are frequent. The HashTree count subsets
		       method does not work for itemsets containing event
		       items. -kap

  July 2nd, 2002     - added special handling of events items with same name
                       -kap

  February 21, 2003  - added weighItemset to count support for a single 
                       itemset. this will be used during candidate generation
		       to remove duplicate itemsets early in an attempt to
		       reduce memory usage, the prohibitive factor right now.
		     - removed data member Vector large. not used since zss
		       incorporated ARMiner into Weka. I don't know why he
		       left it there. -kap

  February 22, 2003  - cleared lists at end of mining method since this
                       is now used iteratively as support is lowered. -kap

  February 24, 2003  - replaced Vector possibleDuplicates with an 
                       ItemsetPrefixTree. this should make looking up duplicate
		       itemsets during generation much much faster... if the
		       overhead of building the tree isn't huge that is.
		     - made consistent the convention of using a lower case
		       letter to start a method name. -kap

  February 27, 2003  - replaced the previous duplicate hash functionality
                       with a lookup in the itemset prefix tree. this is 
		       faster and takes a little less memory. If the 
		       itemset prefix tree is not being used the previous
		       duplicate hash must be used. no mining is done
		       otherwise. -kap

  March 12, 2003     - moved "Beginning to mine..." to 
                       AprioriSetsAndSequences to be consistent -kap

  August 13, 2003    - changed how event items are generated and added to
                       the first level candidate list in findFrequentItemsets.
		       now event items of the same type are numbered 
		       contiguously. this only includes those event items
		       contained in the candidate and frequent itemsets,
		       NOT the data set instances.

  January 19, 2004   - added logger data member for logging metrics -kap
                     - added logging of SECONDS_TO_COUNT_SUPPORT,
		       NUM_FREQUENT_ITEMSETS_FOUND -kap
*/

package wpi.associations.arminerSequence;

import java.util.*;
import java.io.IOException;

// stuff for dealing with time sequence stuff
import wpi.timesequence.*;

/**
 * Implements the AprioriSetsAndSequences version of the 
 * Apriori algorithm for finding frequent itemsets from
 * datasets containing time sequence attributes.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 */

public class ARMinerApriori implements FrequentItemsetsFinder
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** Size to initialize data structures holding itemsets */
  private static final int INITIAL_CAPACITY = 10000;

  /** Use a feature */
  private static final int YES = 1;

  /** Don't use a feature */
  private static final int NO = 0;

  /* List of implementation features to use 
     mostly for performance testing */

  /** Specify use of pruneCandidateGeneration to save calls
   * to getCandidate */
  private static int usePruneCandidateGeneration = YES;

  /** Specify use of itemset prefix tree instead of vector for
      looking up possible duplicate itemsets during generation. */
  private static int useItemsetPrefixTree = YES;
//  private static int useItemsetPrefixTree = NO;

  /** Specify post candidate generation garbage collection.
   Default should be NO */
  private static int usePostGenerationGC = NO;

  /** Specify use of previous duplicate hash for testing possibility
      of an itemset being a duplicate or first of a set of duplicate
      itemsets. This is useful when a large percentage of unique 
      generated itemsets are not duplicates. This can happen when
      there are few sequence event attributes. If set to NO 
      useItemsetPrefixTree must be set to YES. */

  private static int usePreviousDuplicateHash = NO;
//  private static int usePreviousDuplicateHash = YES;

  /** The list of all candidates including those with duplicate
   *  events items. */
  private Vector allCandidates;

  /** The list of candidate itemsets to count support for. */
  private Vector candidates;

  /** The list of frequent itemsets of size k */
  private Vector k_frequent;

  /** The hashtree of candidate itemsets */
  private HashTree ht_candidates;

  /** The hashtree of frequent itemsets of size k */
  private HashTree ht_k_frequent;

  /** The map between duplicate itemsets (containing duplicate event items)
   *  and the first itemset. This is used to update the support of these
   *  itemsets after the support is counted for the others. */
  private Hashtable duplicateHash;

  /** The map between duplicate itemsets (containing duplicate event items)
   *  and the first itemset that were found to be frequent. This is used
   *  for helping identify new generated itemsets as duplicates. */
  private Hashtable previousDuplicateHash = null;

  /* the useItemsetPrefixTree flag controls which of these gets used */

  /** Itemsets that could possibly have duplicates later on. Hopefully
   *  this list will be shorter than the candidate list and hence save time
   *  when searching for the first of a set of duplicate item sets. */
  private Vector possibleDuplicates = null;
  
  /** Itemsets that could possibly have duplicates later on. Hopefully
   *  this list will be shorter than the candidate list and hence save
   *  memory when searching for the first of a set of duplicate item sets.
   *  For larger lists of itemsets this should be much faster than
   *  using the Vector possibleDuplicates. */
  private ItemsetPrefixTree possibleDuplicatesTree = null;

  /** This remembers the number of passes and also indicates the
   *  current cardinality of the candidates. */
  private int pass_num;

  /** Reads rows from a database/dataset file */
  private DBReader db_reader;

  /** Writes itemsets to a cache file */
  private DBCacheWriter cache_writer;
  
  /** Number of rows in the dataset */
  private long num_rows;

  /** Minimum number of rows/instances needed for an itemset to be 
      considered frequent */
  private long min_weight;

  /** The object responsible for logging. */
  protected Logger logger = null;

  // ----- Stuff needed for time sequence attribute support. ----- //

  // the maximum number of items we can add for a single event
  // attribute name. This has the effect of limiting the number of
  // events of the same name can appear in an itemset. 
  // Each event attribute is limited ultimately by the number
  // of items it has in the data set.

  /** Maximum number of events from the same attribute allowed in a rule */
  private int maxEvents;

  // we are only going to use hash trees if we have no
  // event items

  /** Indicates if events are present in the data set. */
  private int haveEvents;

  /** Tracks how many itemsets are pruned using checkSubsets or
      hash tree's count subsets */
  private int numPruned;

  // for debug info -kap

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: and out from method
   * 3: and all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Find the frequent itemsets in a database
   *
   * @param dbReader  
   * the object used to read from the database
   *
   * @param cacheWriter  
   * the object used to write to the cache if this is null,
   * then nothing will be saved, this is useful for
   * benchmarking
   *
   * @param minSupport
   * the minimum support
   *
   * @return  
   * the number of passes executed over the database
   */

  public int findFrequentItemsets ( DBReader dbReader, 
				    DBCacheWriter cacheWriter,
				    float minSupport )
  {
    // check option constraints

    // - if we should NOT use the previous duplicate hash we need
    // - to use the itemset prefix tree for looking up things up
    switch ( usePreviousDuplicateHash )
    {
    case NO:
      switch ( useItemsetPrefixTree )
      {
      case NO:
	System.err.println ( "ARMinerApriori.findFrequentItemsets: " +
			     "usePreviousDuplicateHash set to NO and " +
			     "useItemsetPrefixTree set to NO. " +
			     "We need one of these set to YES." );
	return ( 0 );

      } // end switch use itemset prefix tree

      // if logging, log this
      if (logger != null )
      {
	logger.logValue ( Logger.USE_PREVIOUS_DUPLICATE_HASH, "false" );
      } // end if logging

      break;

    case YES:
      System.out.println ( "ARMinerApriori.findFrequentItemsets: " +
			   "Use previous duplicate hash set to YES. " +
			   "Default is NO." );

      // if logging, log this
      if (logger != null )
      {
	logger.logValue ( Logger.USE_PREVIOUS_DUPLICATE_HASH, "true" );
      } // end if logging

      break;

    } // switch use previous duplicate hash

    // save the following into member fields
    db_reader = dbReader;
    cache_writer = cacheWriter;
    num_rows = dbReader.getNumRows();
    min_weight = (long)( num_rows * minSupport );

    // right now we assume we have no event items
    haveEvents = NO;

    switch ( debug ) {
    case 3:
      System.out.println ( "ARMinerApriori.findFrequentItemsets: " +
			   "make new candidates vector" );
      System.out.println ( "ARMinerApriori.findFrequentItemsets: " +
			   "make new k_frequent vector" );
    } //end switch debug

    // initialize the collections
    candidates = new Vector ( INITIAL_CAPACITY );
    k_frequent = new Vector ( INITIAL_CAPACITY );

    allCandidates = new Vector ( INITIAL_CAPACITY );
    duplicateHash = new Hashtable ( INITIAL_CAPACITY );

    // moved initialization of hash trees to after 1 level generation
    // cause we don't have to at all if we have event items -kap
    //    ht_k_frequent = new HashTree ( k_frequent );
    //    ht_candidates = new HashTree ( candidates );

    // moved this to AprioriSetsAndSequences to be consistent -kap 2003-03-12
    //System.out.println ( "Beginning to mine with min support " + minSupport +
    //			 " confidence " + "...");

    // The initial candidates are all 1-itemsets
    ARMinerItemset is;

    // Time sequence event attributes need to have the number 
    // of desired event items possible in a rule added to the first
    // level of candidates. Each of these items
    // will have relative time "0:1" since there are only two times
    // of concern. This should greatly reduce the initial number
    // of itemsets when the max possible number is low. -kap
 
    // keep track of event attributes we added an item for already -kap
    Vector nameVector = new Vector ( (int)db_reader.getNumColumns() );

    // to track the number we have added so far
    int [] nameCount =  new int [ (int)db_reader.getNumColumns() ];
    
    // have to possibly add items to the number hash... which means we
    // have to add items past the number of columns of our DB -kap
    int newItem = (int)db_reader.getNumColumns() + 1;

    // a counter or two
    int i, ii;

    // should we add the itemset to both the candidates and allCondidates
    // list or just the allCandidates?
    boolean bothCandidates;

    // let's measure the time it takes each support counting and
    // candidate generation phase
    long startTime, endTime, seconds;

    // capture begin time of generating first level of candidates
    // here
    startTime = System.currentTimeMillis();

    for ( i = 1; i <= db_reader.getNumColumns(); i++ ) 
    {
      // and we'll want to add it to both candidate lists
      bothCandidates = true;

      // we must check if the item i is an event -kap
      // get the value of item i
      Object o = db_reader.getNumberHash().get ( new Integer ( i ) );

      // now we have to seperate the value from the attribute
      // value pair
      StringTokenizer st = new StringTokenizer ( (String)o, "=" );
      
      String eventName = "";

      if ( st.hasMoreTokens() )
      {
	// keep first token, if this turns out to be
	// a time sequence event we will want to put this
	// in the name hash
	eventName = st.nextToken();
      }

      String s ="";

      if ( st.hasMoreTokens() )
      {
	// get the token after =
	s = st.nextToken();
      }

      // changing this so that time sequence events are stored
      // until all the regular events are taken care of.
      // then the event items are created and we can assume
      // that event items of the same name/type are numbered
      // contiguously. this will only apply to event items
      // in the candidate and frequent itemsets. -kap 2003-08-13

      // is it a time sequence event?
      if ( TSUtility.isTimeSequenceEvent ( s ) )
      {
	// we must check if we have already added an item for
	// that event attribute. -kap
	if ( nameVector.contains ( eventName ) )
	{
	  // then increment the count
	  nameCount [ nameVector.indexOf ( eventName ) ]++;

	} // end if we already have an item for this event attribute
	else
	{
	  // add the name to our name list
	  nameVector.addElement ( eventName );

	  // increment the name count -kap 2003-10-02
	  nameCount [ nameVector.indexOf ( eventName ) ]++;
	  
	  // and make sure we know there are events
	  haveEvents = YES;

	} // end else we have to add the name to the name hash

	// well, we can skip the rest for item i (the event item)
	// and continue with the next item
	continue;

      } // end if item i is a time sequence event

      is = new ARMinerItemset ( 1 );
	
      // we must pass the number hash from the DBReader
      // to the new itemset to support time sequence
      // attributes -kap
      is.setNumberHash ( db_reader.getNumberHash() );
      
      is.addItem ( i );

      // simply add the item to both lists
      allCandidates.add ( is );
      candidates.add ( is );

      switch ( debug ) {
      case 3: //case 0:
	System.out.println ( "ARMinerApriori.findFrequentItemsets: " +
			     "Added itemset = " + is.toString() + " " +
			     "with item " + s + " to candidates." );
      } // end switch debug -kap

      // we don't have to worry about adding the item to both lists
      // since we are not dealing with event items yet. only event
      // items will have duplicates. -kap 2003-08-13

      // moved this to after the generation loop cause we don't
      // have to at all if there are event items -kap

      //      ht_candidates.add ( candidates.size() - 1 );
      
    } // end for all the level 1 candidates.

    // ok, so now we have left over a list of event names
    // and the number of occurances of that name in our data set
    // we have to make an event item for each occurrance, not exceeding
    // the maximum number of duplicate event items allowed.

    // go through the name vector and make the event items

    // the number of event items to make with a certain name
    int n, c;

    for ( i = 0; i < nameVector.size(); i++ )
    {
      // how many to make?
      n = Math.min ( nameCount [ i ], maxEvents );

      for ( c = 0; c < n; c++ )
      {
	// add the new item to the number hash with value begin time
	// 0 and end time 1
	// first build the value... using string buffer should be a little
	// quicker
	StringBuffer sb = new StringBuffer ( (String)(nameVector.get ( i )) );
	sb.append ( "=0" );
	sb.append ( TSUtility.TIME_SEQUENCE_DELIMITER );
	sb.append ( "1" );
	db_reader.getNumberHash().put ( new Integer ( newItem ),
					sb.toString() );
	switch ( debug ) {
	case 3: //case 0:
	  System.out.println ( "ARMinerApriori.findFrequentItemsets: " +
			       "Added event = " + sb.toString() + " " +
			       "to number hash with item # " + newItem );
	} // end switch debug -kap

	// have to make the candidate itemset
	is = new ARMinerItemset ( 1 );
	
	// we must pass the number hash from the DBReader
	// to the new itemset to support time sequence
	// attributes -kap
	is.setNumberHash ( db_reader.getNumberHash() );
	
	is.addItem ( newItem );

	// if this is the first event item of this type
	// add it to both lists
	switch ( c )
	{
	case 0:
	  // add to both lists
	  allCandidates.add ( is );
	  candidates.add ( is );
	  break;

	default:
	  // add to all list and add to duplicate hash
	  // the original candidate is the last in the list
	  // since we must have just added it
	  duplicateHash.put ( is, candidates.get ( candidates.size() - 1 ) );

	  allCandidates.add ( is );
	  break;
	  
	} // end switch c
	
	// increment new item so we'll be ready for the next one
	newItem++;

      } // end for each event item for this name

    } // end for event names

    // if we have no events, initialize the hash tree stuff
    switch ( haveEvents )
    {
    case NO:
      System.out.println ( "There are no event items in the data set " +
			   "so we'll use hash trees." );
      // initialize the hash trees
      ht_k_frequent = new HashTree ( k_frequent );
      ht_candidates = new HashTree ( candidates );
      
      // add the first generation candidates to the hash tree
      for ( i = 0; i < candidates.size(); i++ )
      {
	ht_candidates.add ( i );

      } // end for each candidate

      break;

    case YES:
      // if we are using the vector for duplicates say so since
      // we should usually use the prefix tree
      switch ( useItemsetPrefixTree )
      {
      case NO:
	System.out.println ( "Using Vector for possible duplicate itemsets. " +
			     "Should use Itemset Prefix Tree, much faster. " +
			     "You must be testing this. How clever you are!" );

	// if logging, log this
	if (logger != null )
	{
	  logger.logValue ( Logger.USE_ITEMSET_PREFIX_TREE, "false" );
	} // end if logging

	break;

      case YES:
	// if logging, log this
	if (logger != null )
	{
	  logger.logValue ( Logger.USE_ITEMSET_PREFIX_TREE, "true" );
	} // end if logging

	break;

      } // end switch use itemset prefix tree

    } // end switch have events

    // capture end time of generating first level candidates here
    // and log if necessary
    endTime =  System.currentTimeMillis();

    // calculate how many seconds went by
    seconds = ( endTime / 1000 ) - ( startTime / 1000 );

    if ( seconds > 1 )
    {
      System.out.print ( seconds + " seconds: " );

    } // end if more than 1 second
    else
    {
      System.out.print ( seconds + " second: " );

    } // end else 1 or less seconds

    System.out.println ( "Level 1 candidates: " + 
			 allCandidates.size() +
			 " (" + candidates.size() + ")" );

    if ( logger != null )
    {
      logger.logValue ( Logger.SECONDS_TO_GENERATE_CANDIDATES,
			Long.toString ( seconds ) );

      logger.logValue ( Logger.NUM_CANDIDATES_GENERATED, 
			Integer.toString ( allCandidates.size() ) );

      logger.logValue ( Logger.NUM_UNIQUE_CANDIDATES_GENERATED,
			Integer.toString ( candidates.size() ) );

    } // end if we are logging

    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "ARMinerApriori.findFrequentItemsets: " +
			   "duplicateHash= " + duplicateHash );
    } // end switch debug

    // we start with first pass
    for ( pass_num = 1; ; pass_num++ )
    {
      // we want to see how long it takes to count support so
      // record the time we start
      startTime = System.currentTimeMillis();

      // compute the weight of each candidate
      weighCandidates();

      // verify which candidates are frequent (have weight
      // greater than or equal to min_weight)
      evaluateCandidates();

      // and the time we stop
      endTime =  System.currentTimeMillis();

      // calculate how many seconds went by
      seconds = ( endTime / 1000 ) - ( startTime / 1000 );
      
      if ( seconds > 1 )
      {
	System.out.print ( seconds + " seconds: " );

      } // end if more than one second
      else
      {
	System.out.print ( seconds + "  second: " );

      } // end else one second or less

      System.out.println ( "# of frequent itemsets for level " + pass_num
			   + " = " + k_frequent.size() );

      // log stuff if we have a logger
      if ( logger != null )
      {
	logger.logValue ( Logger.SECONDS_TO_COUNT_SUPPORT, 
			  Long.toString ( seconds ) );

	logger.logValue ( Logger.NUM_FREQUENT_ITEMSETS_FOUND,
			  Integer.toString ( k_frequent.size() ) );
	
      } // end if we are logging

      // if there are no frequent itemsets we can stop mining
      if ( k_frequent.size() == 0 )
      {
	break;
      }

      // if we just examined the top itemset (the one containing
      // all items) then we're done, nothing more to do. -zss
      if ( pass_num >= db_reader.getNumColumns() )
      {
	break;
      }

      // we want to see how long it takes to generate candidates so
      // record the time we start
      startTime = System.currentTimeMillis();

      // generate new candidates from frequent itemsets
      generateCandidates();

      // and record the time we stop
      endTime = System.currentTimeMillis();

      // calculate how many seconds went by
      seconds = ( endTime / 1000 ) - ( startTime / 1000 );

      // report the time to standard out
      if ( seconds > 1 )
      {
	System.out.print ( seconds + " seconds: " );

      } // end if more than one second
      else
      {
	System.out.print ( seconds + "  second: " );

      } // end else one second or less

      System.out.println ( "# of Generated candidates for level: " 
			   + ( pass_num + 1 ) + " = " + allCandidates.size() +
			   " (" + candidates.size() + ")" );

      // log stuff if logging is on
      if ( logger != null )
      {
	logger.logValue ( Logger.SECONDS_TO_GENERATE_CANDIDATES,
			  Long.toString ( seconds ) );

	logger.logValue ( Logger.NUM_CANDIDATES_GENERATED,
			  Integer.toString ( allCandidates.size() ) );

	logger.logValue ( Logger.NUM_UNIQUE_CANDIDATES_GENERATED,
			  Integer.toString ( candidates.size() ) );

      } // end if logging on

      // exit if no more candidates
      if ( candidates.size() == 0 )
      {
	// if we break here we didn't have to count support of the
	// candidates, let's log that so we have the same amount
	// of log info for generation and support counting
	if ( logger != null )
	{
	  logger.logValue ( Logger.SECONDS_TO_COUNT_SUPPORT, "0" );
	  logger.logValue ( Logger.NUM_FREQUENT_ITEMSETS_FOUND, "0" );
	  logger.logValue ( Logger.NUM_UNIQUE_FREQUENT_ITEMSETS_FOUND, "0" );
	  logger.logValue ( Logger.NUM_COMPARISONS_SAVED_REMOVING_CANDIDATES_DURING_SUPPORT_COUNTING, "0" );

	} // end if logging

	break;
      } // end if no candidates left

    } // end for each pass

    // it is important to clean up our lists and things since
    // this method may be used iteratively as support is lowered
    // if this round did not produce any usable rules -kap
    candidates.clear();
    candidates.trimToSize();
    k_frequent.clear();
    k_frequent.trimToSize();

    allCandidates.clear();
    allCandidates.trimToSize();
    duplicateHash.clear();

    return ( pass_num );

  } // END public int findFrequentItemsets ( DBReader dbReader, 
  //				 DBCacheWriter cacheWriter,
  //				 float minSupport )

//********************************************************//

  /**
   * This procedure scans the database and computes the
   * weight of each candidate.
   */

  private void weighCandidates()
  {
    // print out status of counting supports, need the .0
    // so we get a double back...otherwise when num rows less
    // than 50 we get print status = 0 and that's no good -kap
    int printStatus = (int)Math.ceil ( db_reader.getNumRows() / 50.0 );

    System.out.println ( "ARMinerApriori.weighCandidates: " +
			 "counting supports: " + db_reader.getNumRows() + 
			 " rows, 1:" + printStatus + " - " );

    // since it seems really icky to make hashtree 
    // (which ht_candidates is) work with event items we are just
    // going to do the updating of weight brute force over the
    // vector of itemsets -kap
    
    // actually to keep processing data sets not containing event items
    // efficient we use hash trees when there are no event items -kap
    switch ( haveEvents )
    {
    case NO:
      ht_candidates.prepareForDescent();
      
    } // end switch have events
    
    // rearranged try so that only exceptions from the
    // db_reader would be caught by this method -kap
    ARMinerItemset row = null;

    try
    {
      row = db_reader.getFirstRow();
    } // end try
    catch ( Exception e )
    {
      System.err.println ( "ARMinerApriori.weighCandidates : " +
			   "Error scanning database!!!\n" + e );
    }

    int index;
    ARMinerItemset itemset;

    // to keep processing data sets not containing event items
    // efficient we use hash trees when there are no event items -kap
    switch ( haveEvents )
    {
    case NO:
      ht_candidates.update ( row );
      break;

    case YES:
      for ( index = 0; index < candidates.size(); index++ )
      {
	itemset = (ARMinerItemset)candidates.get ( index );

	if ( itemset.isIncludedIn ( row ) )
	{
	  itemset.incrementWeight();

	} // end if itemset is included in row

	// print out status of counting supports
	switch ( debug ) {
	case 1:
	  System.out.print ( ":" );
	} // end switch debug
	
      } // end for each candidate
      break;

    default:
      // this should never happen
      System.err.println ( "ARMinerApriori.weighCandidates: " +
			   "neither events nor no events. Weird." );
    } // end switch have events

    // print out status of counting supports
    switch ( printStatus ) 
    {
    case 1: // only print a | if we're doing 1 for 1
      System.out.print ( "|" );

    } // end switch print status

    // count the rows
    int numRow = 1;

    // count the number of itemset comparisons we save by removing
    // candidates that will never reach min support
    int savedComparisons = 0;
    long possibleWeight;

    while ( db_reader.hasMoreRows() )
    {
      try
      {
	row = db_reader.getNextRow();
      } // end try
      catch ( Exception e )
      {
	System.err.println ( "ARMinerApriori.weighCandidates : " +
			     "Error scanning database!!!\n" + e );
      }
      
      // since it seems really icky to make hashtree 
      // (which ht_candidates is) work with event items we are just
      // going to do the updating of weight brute force over the
      // vector of itemsets -kap

      // actually to keep processing data sets not containing event items
      // efficient we use hash trees when there are no event items -kap
      switch ( haveEvents )
      {
      case NO:
	ht_candidates.update ( row );
	break;
      case YES:
	for ( index = 0; index < candidates.size(); index++ )
	{
	  itemset = (ARMinerItemset)candidates.get ( index );

	  if ( itemset.isIncludedIn ( row ) )
	  {
	    itemset.incrementWeight();
	    
	  } // end if itemset is included in row
	  
	  // print out status of counting supports
	  switch ( debug ) {
	  case 2: //case 0:
	      System.out.println ( "ARMinerApriori.weighCandidates: " +
				   "item set: " + itemset + 
				   " included in instance: " + row );
	      break;
	  case 1:
	    System.out.print ( ":" );
	  } // end switch debug

	  // check if it is possible for this itemset to meet the weight
	  // requirement.
	  
	  // check how many rows are left in the data set
	  // if this plus the weight will meet the min weight
	  // keep it, if not, throw it and its dups out
	  possibleWeight = itemset.getWeight() + 
	    db_reader.getNumRows() - ( numRow + 1 );

	  if ( possibleWeight < min_weight )
	  {
	    // remove it from the candidate list since we don't have to
	    // bother counting it any more
	    candidates.removeElementAt ( index );

	    // add the number of comparisons we saved by removing the itemset
	    savedComparisons += db_reader.getNumRows() - ( numRow + 1 );

	    // and set index back 1 so we don't skip the next itemset
	    index--;

	    switch ( debug ) {
	    case 3: //case 0:
	      System.out.println ( "ARMinerApriori.weighCandidates: " +
				   "removed itemset " + itemset + " " +
				   "from candidate list after row " + 
				   ( numRow + 1 ) + " Possible weight: " +
				   possibleWeight + " Min: " + min_weight );
	    } // end switch debug

	  } // end if itemset won't make min weight
		  
	} // end for each candidate
	break;

      } // end switch have events

      // print out status of counting supports
      switch ( (int)( ( numRow + 1 ) % printStatus ) ) 
      {
      case 0:
	System.out.print ( "|" );
	
      } // end switch print status
      
      // increment row count
      numRow++;

    } // end while more rows
    
    switch ( debug ) {
    case 0:
      System.out.println ( "" );
    } // end switch debug

    // now update the weights of the duplicate items

    // get the list of duplicates
    Enumeration dups = duplicateHash.keys();
    ARMinerItemset dup;
    
    while ( dups.hasMoreElements() )
    {
      dup  = (ARMinerItemset)dups.nextElement();

      dup.setWeight 
	( ((ARMinerItemset)duplicateHash.get ( dup )).getWeight() );

      // don't need this here since we moved event weight counting
      // to until after rule generation -kap 2004-01-19
      //      dup.setEventWeight 
      //      ( ((ARMinerItemset)duplicateHash.get ( dup )).getEventWeight() );

    } // end while more duplicates

    switch ( debug ) {
    case 3:
      System.out.println ( "ARMinerApriori.weighCandidates: " +
			   "duplicateHash= " + duplicateHash );
    case 2:
      System.out.println ( "ARMinerApriori.weighCandidates: " +
			   "removing candidates saved " + 
			   savedComparisons + " comparisons." );
    } // end switch debug

    if ( logger != null )
    {
      logger.logValue ( Logger.NUM_UNIQUE_FREQUENT_ITEMSETS_FOUND,
			Integer.toString ( candidates.size() ) );
      
      logger.logValue 
	( Logger.
	  NUM_COMPARISONS_SAVED_REMOVING_CANDIDATES_DURING_SUPPORT_COUNTING,
	  Integer.toString ( savedComparisons ) );
      
    } // end if logging on

  } // END private void weighCandidates()

//********************************************************//

  /**
   * This procedure scans the database and computes the
   * weight of the itemset.
   */

  private void weighItemset ( ARMinerItemset itemset )
  {
    // try so that only exceptions from the
    // db_reader would be caught by this method
    ARMinerItemset row = null;

    try
    {
      row = db_reader.getFirstRow();
    } // end try
    catch ( Exception e )
    {
      System.err.println ( "ARMinerApriori.weighItemset : " +
			   "Error scanning database!!!\n" + e );
    }

    if ( itemset.isIncludedIn ( row ) )
    {
      itemset.incrementWeight();

    } // end if itemset is included in row
    
    // count the rows
    int numRow = 1;
    
    while ( db_reader.hasMoreRows() )
    {
      try
      {
	row = db_reader.getNextRow();
      } // end try
      catch ( Exception e )
      {
	System.err.println ( "ARMinerApriori.weighItemset : " +
			     "Error scanning database!!!\n" + e );
      }
	
      if ( itemset.isIncludedIn ( row ) )
      {
	itemset.incrementWeight();
	  
      } // end if itemset is included in row
	
    } // end while db has more rows

  } // END private void weighItemset ( ARMinerItemset itemset )

//********************************************************//

  /**
   * This procedure checks to see which itemsets are frequent
   */

  private void evaluateCandidates()
  {
    ARMinerItemset is;

    // this has been changed to use the allCandidates list instead
    // of candidates. -kap
    
    for ( int i = 0; i < allCandidates.size(); i++ )
    {
      // if this is a frequent itemset
      if ( ( is = (ARMinerItemset)allCandidates.get ( i ) ).getWeight() >= 
	   min_weight )
      {
	// compute support of itemset
	is.setSupport ( (float)is.getWeight() / (float)num_rows );
	
	// write itemset to the cache
	try
	{
	  if ( cache_writer != null )
	  {
	    switch ( debug ) {
	    case 3: case 2: //case 0:
	      System.out.println ( "ARMinerApriori.evaluateCandidates: " +
				   "writing itemset = " + is );
	    } // end swith debug -kap

	    cache_writer.writeItemset ( is );
	  }
	} // end try
	catch ( IOException e )
	{
	  System.err.println (  "ARMinerApriori.evaluateCandidates: " +
				"Fishy IO error!\n" + e );
	} // end catch

	// then add it to the large and k_frequent collections
	switch ( debug ) {
	case 3: case 2: //case 0:
	  System.out.println ( "ARMinerApriori.evaluateCandidates: " +
			       "adding itemset = " + is + " " +
			       "to frequent collection." );
	} // end switch debug -kap

	// ----- Ya know, I noticed that there is no code in ARMiner
	// ----- nor Weka that closes the cache file. That's kinda
	// ----- bad, right? I also noticed that there are no allowances
	// ----- for multiple instances of this to be running. If a file
	// ----- temp.cache is in use, not that that is checked anyway,
	// ----- it seems to be simply overwritten which would cause
	// ----- all sorts of problems. -kap

	k_frequent.add ( is );

	// only use hash tree if we have no events
	switch ( haveEvents )
	{
	case NO:
	  ht_k_frequent.add ( k_frequent.size() - 1 );

	} // end switch have events

      } // end if frequent itemset

    } // end for itemsets in allCandidates list

    // reinitialize candidates for next step
    candidates.clear();

    // remember to clear the all candidates list too
    allCandidates.clear();

    // make a new duplicate hash
    // and save the old one if that option is specified
    switch ( usePreviousDuplicateHash )
    {
    case YES:
      previousDuplicateHash = duplicateHash;
      break;
    } // end switch previous duplicate hash

    duplicateHash = new Hashtable();

    // only use hash tree if we have no events
    switch ( haveEvents )
    {
    case NO:
      ht_candidates = new HashTree ( candidates );

    } // end switch have events

    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "ARMinerApriori.evaluateCandidates: " +
			   "duplicateHash= " + duplicateHash );
    } // end switch debug

  } // END private void evaluateCandidates()

//********************************************************//

  /**
   * Generates new candidates out of itemsets that are
   * frequent.
   */

  private void generateCandidates()
  {
    // need frequent itemsets to generate candidates
    if ( k_frequent.size() == 0 )
    {
      return;
    } // end if no frequent itemsets

    // only use hash tree if we have no events
    switch ( haveEvents )
    {
    case NO:
      ht_k_frequent.prepareForDescent();
      break;

    case YES:
      // make a new possible duplicate
      switch ( useItemsetPrefixTree )
      {
      case NO:
	possibleDuplicates = new Vector();
	break;
      case YES:
	possibleDuplicatesTree = new ItemsetPrefixTree();
	break;
      } /// end switch use itemset prefix tree

    } // end switch have events

    // reset the pruned counter so we know how many itemsets we prune
    // for this batch of candidates using checkSubsets
    numPruned = 0;

    // count the number of getCandidate calls we save by discontinuing
    // after a failed check regular items
    int savedCalls = 0;

    // itemsets we will work with
    ARMinerItemset isi, isj;
      
    // just wanna see if I would lose any itemsets by using the pruning below
    // only need this if reporting and that requires code changes below.
    // just uncomment if you wanna do it -kap
    //    boolean prune = false;

    // for each frequent itemset try combining it with the other
    // frequent itemsets to generate candidates
    for ( int i = 0; i < k_frequent.size() - 1; i++ ) 
    {
      // print out a character when we start generating candidates
      // using the next batch of frequent itemsets
      if ( ( i % 100 ) == 0 )
      {
	System.out.print ( '`' );

	// might want to report the candidates generated so far
	switch ( debug ) {
	case 0:
	  System.out.print ( "(" + allCandidates.size() + "/" +
			     candidates.size() + ")" );
	} // end switch debug
      }

      for ( int j = i + 1; j < k_frequent.size(); j++ ) 
      {
	// get the itemsets we will work with
	isi = (ARMinerItemset)k_frequent.get ( i );
	isj = (ARMinerItemset)k_frequent.get ( j );

	// only use hash tree if we have no events
	// use and if else here instead of switch so we can break
	// out of the loop if we have no events and getting a candidate
	// returns false
	if ( haveEvents == YES )
	{
	  // lets compare the regular items first.
	  // if they do not match then we can break and go on to
	  // the next i itemset
	  // luckily there's a method to just tell us if we should
	  // prune this line of candidate generation
	  if ( pruneCandidateGeneration ( isi, isj ) )
	  {
	    // might want to verify that we are pruning correct
	    // itemset generation
	    switch ( debug ) {
	    case 3:
	      System.out.println ( "ARMinerApriori.generateCandidates: " +
				   "pruneCandidateGeneration: " +
				   pruneCandidateGeneration ( isi, isj ) +
				   " for itemsets: \n" + isi + "\n" +
				   isj );
	    } // end switch debug

	    // count how many calls to get candidate we save
	    savedCalls += k_frequent.size() - j;

	    break;

	  } // end if we should prune the rest of candidates for i
	  else
	  {
	    getCandidate ( isi, isj );

	  } // end else there's still some good candidates left to generate
	  // keep on trucking

	} // end we have events
	else
	{
	  // if candidate generation fails stop with i and go to i+1
	  if ( ! getCandidate ( isi, isj ) )
	  {
	    break;
	    
	    // now, this break is here because...
	    
	    //   from the getCandidate method:
	    //   if we cannot combine element i with j then we shouldn't
	    //   waste time for bigger j's. This is because we keep the
	    //   collections ordered, an important detail in this
	    //   implementation
	    
	    // the problem is with event items. the number of the
	    // item does not matter in determining if itemsets can
	    // be combined. 
	    
	    // I should learn more about how the collections are ordered
	    // so I can some how use the pruning. Right now it has been
	    // empirically shown that valid candidates containing event
	    // items do get pruned. To see this uncomment the pruning code
	    // and set the debug level to 3 (there will be a lot of info)
	    // or change the case to 0. -kap

	    // it looks like we can still break if the first item in
	    // the ith itemset is NOT an event, yay!
	    
	    /* // uncomment this to see wrongly pruned itemsets... you'll 
	    // also have to change stuff to use the hash trees all the time...
	    // which would result in everything not working really so
	    // you'll have to take my word for it
	    
	    // in the meantime make note of the prune
	    prune = true;
	    
	    switch ( debug ) {
	    case 0:
	      System.out.println ( "ARMinerApriori.generateCandidates: " +
				   "prune combining itemsets i= " + i + " - " +
				   k_frequent.size() + ", j= " + j + " - " +
				   k_frequent.size() );
	    } // end switch debug

	    // uncomment this to see wrongly pruned itemsets
	    
	  } // end if we could not combine i and j to get a new candidate
	  else
	  {
	    // this is just for my own curiosity
	    if ( prune )
	    {
	      System.err.println ( "ARMinerApriori.generateCandidates: " +
				   "itemset combining i= " + i + " and " + j +
				   " would have been pruned incorrectly: " +
				   candidates.lastElement() );
	    } // end if we would have pruned this itemset
	    
	  } // end else we could combine i and j

	  */ // uncomment this and comment out the next line
	     // to see wrongly pruned itemsets

	  } // end if getting candidate didn't work

	} // end else there are no event items

      } // end for each other frequent itemset

    } // end for each frequent itemset
    
    switch ( debug ) {
    case 3: //case 0:
      System.out.print ( "\nARMinerApriori.generateCandidates: " +
			 "pruned " + numPruned + " " );
      switch ( haveEvents ) {
      case NO:
	System.out.print ( "itemsets using hash tree count subsets.\n" );
	break;
      case YES:
	System.out.print ( "itemsets using checkSubsets.\n" );

	System.out.print ( "Possible first of duplicates: " );
	
	switch ( useItemsetPrefixTree )
	{
	case NO:
	  System.out.print ( possibleDuplicates.size() + ".\n" );
	  break;
	case YES:
	  System.out.print ( possibleDuplicatesTree.size() + ".\n" );
	} // end switch use itemset prefix tree

      } // end switch haveEvents

      //    case 0: // uncomment if we remove the case 0 from above
      // we might want to report how many calls to getCandidate
      // we saved... if any were saved
      switch ( savedCalls ) {
      case 0:
	break;
	
      default:
	System.out.println ( "ARMinerApriori.generateCandidates: " +
			     "Calls to getCandidate saved: " + savedCalls );
	} // end switch saved calls

    } // end switch debug

    // reinitialize k_frequent for next step
    k_frequent.clear();

    // only use hash tree if we have no events
    switch ( haveEvents )
    {
    case NO:
      ht_k_frequent = new HashTree ( k_frequent );
      break;

    case YES:
      // we don't need the possible duplicates list any more -kap 2003-02-21
      possibleDuplicates = null;
      possibleDuplicatesTree = null;

    } // end switch have events

    // let's try collecting some garbage, this may take
    // a little while but after generation we may have
    // a lot of memory to reclaim... so we'll make it an
    // option
    switch ( usePostGenerationGC )
    {
    case YES:
      System.gc();
      break;
    } // end switch use post generation garbage colection

    switch ( debug ) {
    case 3: // case 0:
      System.out.print ( "\nARMinerApriori.generateCandidates: " +
			 "generated candidates: " + allCandidates );
    } // end switch debug

  } // END private void generateCandidates()

//********************************************************//

  /**
   * This procedure tries to combine itemsets i and j and
   * returns true if succesful, false if it can't combine
   * them.
   */

  private boolean getCandidate ( int i, int j )
  {
    return ( getCandidate ( (ARMinerItemset)k_frequent.get ( i ),
			    (ARMinerItemset)k_frequent.get ( j ) ) );

  } // END private boolean getCandidate ( int i, int j )

//********************************************************//

  /**
   * This procedure tries to combine itemsets i and j and
   * returns true if succesful, false if it can't combine
   * them.
   */

  private boolean getCandidate ( ARMinerItemset is_i, 
				 ARMinerItemset is_j )
  {
    // if we cannot combine element i with j then we shouldn't
    // waste time for bigger j's. This is because we keep the
    // collections ordered, an important detail in this implementation

    // to support time sequence stuff canCombineWith now returns
    // a hastable mapping event labels. If it is null then the
    // itemsets cannot be combined. This hashtable is now required
    // by the ARMinerItemset.combineWith method. -kap

    Hashtable eventLabelMap = is_i.canCombineWith ( is_j );

    if ( eventLabelMap == null )
    {
      switch ( debug ) {
      case 3: //case 0:
	System.out.println ( "ARMinerApriori.getCandidate: " +
			     "could not combine: " + is_i + " with events= " +
			     is_i.getEventVector() + " and " +
			     is_j + " with events= " + is_j.getEventVector() );
      } // end switch debug

      return ( false );
    } 

    // to support time sequences combineWith has been changed to
    // return a list of itemsets and take a hashtable mapping
    // event labels as an argument -kap
    
    //      ARMinerItemset is = is_i.combineWith ( is_j );
    Enumeration e = is_i.combineWith ( is_j, eventLabelMap ).elements();
    ARMinerItemset is, isDup;

    // so now we add each itemset in the list -kap
    while ( e.hasMoreElements() )
    {
      is = (ARMinerItemset)e.nextElement();
      
      // a real k-frequent itemset has k (k-1)-frequent subsets

      // ----- start kap commentary ----- //
      // well, this may be true for itemsets containing only
      // regular items but if an itemset has more than one time
      // sequence event there may be more subsets than that,
      // since the relative times can differ. A simple fix would be
      // to change == to >= but that would be invalid for some itemsets.
      // What this is really trying to accomplish is making sure that
      // all the subsets of the itemset "is" of size k-1 are frequent.
      // Unfortunately this was done by counting the subsets in the hash
      // tree "ht_k_frequent" that are included in "is". With itemsets
      // containing events there are more so the count is higher and the
      // condition below fails even though "is" is a valid and frequent
      // itemset.
      // changing == to >= for now even though some itemsets may
      // not get pruned... but our results will still be valid.
      
      // now I'm going back to the original == and only using it
      // and hash tress at all, if there are no events in the data set.
      // if there are events the checkSubsets method is used. -kap
      // ----- end kap commentary ----- //

      switch ( haveEvents )
      {
      case NO: // have events

	if ( ht_k_frequent.countSubsets ( is ) == is.size() )
	{
	  // we always add the itemset to the all candidates
	  // list
	  allCandidates.add ( is );

	  candidates.add ( is );
	  ht_candidates.add ( candidates.size() - 1 );
	  
	} // end if subset count correct
	else
	{
	  numPruned++;
	} // end else prune the itemset

	break;
	
      case YES: // have events

	// check if each subset is also frequent... and yes, it does
	// it differently for itemsets with events
	if ( checkSubsets ( is ) )
	{
	  // we always add the itemset to the all candidates
	  // list
	  allCandidates.add ( is );

	  // check if this is a duplicate itemset...
	  // we may save some time by checking if itemsets i or j
	  // are duplicates first if the option is specified
	  switch ( usePreviousDuplicateHash )
	  {
	  case YES: // use previous duplicate hash
	    
	    if ( previousDuplicateHash.containsKey ( is_i ) ||
		 previousDuplicateHash.containsKey ( is_j ) )
	    {
	      // check the candidate set... if "is" is a dup itemset
	      // we should find it in the possible duplicate list or tree
	      
	      switch ( useItemsetPrefixTree )
	      {
	      case NO:

		// better search the possible dups we have all ready 
		int ii;
		for ( ii = possibleDuplicates.size() - 1; ii > -1; ii-- )
		{
		  if ( ((ARMinerItemset)possibleDuplicates.
			get ( ii )).isIncludedIn ( is ) )
		  {
		    // add this to the duplicate hash
		    duplicateHash.put ( is, possibleDuplicates.get ( ii ) );
		    break;
		    
		  } // end if we found duplicate
		  
		} // end for possibleDuplicates so far, going backwards
		
		// add the itemset to candidates if we didn't find a duplicate
		if ( ii < 0 )
		{
		  candidates.add ( is );
		  
		} // end if we should add the itemset to candidate (not a dup)
		break;
		
	      case YES: // use itemset prefix tree

		// is there a duplicate of itemset "is"
		isDup = possibleDuplicatesTree.getItemset ( is );
		
		if ( isDup == null )
		{
		  // then we just add this to the candidates list
		  candidates.add ( is );
		  
		} // end if no dup
		else
		{
		  // add the dup to the duplicate hash then
		  duplicateHash.put ( is, isDup );
		  
		} // end else yes dup
		break;
		
	      } // end switch use itemset prefix tree
	      
	    } // end if i or j is a duplicate
	    else
	    {
	      candidates.add ( is );

	      if ( previousDuplicateHash.containsValue ( is_i ) ||
		   previousDuplicateHash.containsValue ( is_j ) )
	      {
		// add is to the possible list then
		switch ( useItemsetPrefixTree )
		{
		case NO:
		  possibleDuplicates.add ( is );
		  break;
		case YES:
		  possibleDuplicatesTree.add ( is );
		  break;
		} // end switch use itemset prefix tree
		
	      } // end if i or j was the first of a set of dups
	      
	    } // end else it isn't a dup
	    
	    break;
	    
	  case NO: // use previous duplicate hash

	    // then we can assume that we are using the itemset prefix tree
	    // cause find associations checks for the condition first

	    // if it is a dup then we should find it in the prefix tree
	    isDup = possibleDuplicatesTree.getItemset ( is );

	    if ( isDup == null )
	    {
	      // then we just add this to the candidates list
	      candidates.add ( is );

	      // and to the possible duplicates tree... since we are
	      // forced to assume it could have duplicates since
	      // we cannot check in the previous duplicates hash
	      possibleDuplicatesTree.add ( is );	      

	    } // end if no dup
	    else
	    {
	      // add the dup to the duplicate hash then
	      duplicateHash.put ( is, isDup );
	      
	    } // end else yes dup

	    break; // end case NO

	  } // end switch use previous duplicate hash

	  } // end if all the subsets are found
	else
	{
	  // just curious how many we will prune
	  numPruned++;

	  switch ( debug ) {
	  case 3: //case 0:
	    System.out.println ( "ARMinerApriori.getCandidate: " +
				 "pruned itemset: " + is );
	  } // end switch debug

	} // end else we pruned the "is" itemset

	break;
	
      } // end switch have events

      } // end while e has more -kap

    return ( true );

  } // END private boolean getCandidate ( ARMinerItemset is_i, 
  //				 ARMinerItemset is_j )

//********************************************************//

  /**
   * Checks to see if all the subsets of the specified
   * itemset are frequent.
   *
   * @param itemset
   * the itemset to check the subsets of
   *
   * @return true if all the subsets are frequent, 
   * false otherwise
   */

  private boolean checkSubsets ( ARMinerItemset itemset )
  {
    // we want to do all but the two last subsets.
    // we don't have to test the ones without either the
    // two last items since we know they are frequent already.
    // how do we know? well, I am assuming that the only
    // placed this is called from is the getCadidate method.
    // getCandidate just used those two itemsets to create
    // the one we are testing now... nifty, huh? -kap

    // before we do anything, if itemset is size 2 or less
    // we can skip checking all together. If you're wondering
    // why read above again.

    if ( itemset.size() < 3 )
    {
      return ( true );
    } // end if itemset so small we know already

    // the subset to look for in frequent itemset list
    ARMinerItemset subset;

    // the item to remove from itemset to create subset
    int i, f;
    
    // remember to skip the last 2 items, these must be in each subset
    // that gets checked
    for ( i = 0; i < itemset.size() - 2; i++ )
    {
      // make a copy of the itemset and remove the i-th item
      subset = new ARMinerItemset ( itemset );
      subset.removeItem ( itemset.getItem ( i ) );

      // now look for subset in the k_frequent itemset list
      for ( f = 0; f < k_frequent.size(); f++ )
      {
	// This iteration through the frequent items from the last level 
	// could be a great deal faster... maybe using SET to retrieve 
	// the support. -kap
	if ( subset.isIncludedIn ( (ARMinerItemset)k_frequent.get ( f ) ) )
	{
	  // just stop looking
	  break;
	} // end if we found the subset is frequent 

      } // end for each frequent itemset of size k 

      // if we had to exhaust our list looking then is isn't frequent
      if ( f == k_frequent.size() )
      {
	return ( false );

      } // end if we didn't find the subset was frequent

    } // end for all but the last two subsets

    // well, we found 'em all I guess
    return ( true );

  } // END private boolean checkSubsets ( ARMinerItemset itemset )

//********************************************************//

  /**
   * This method is used to prune candidate generation when
   * the dataset contains events. It is assumed that the
   * itemsets are of the same length.
   * Compares only the regular items in the specified 
   * itemsets. Returns true if the generation of candidates
   * should be stopped with the current isi itemset.
   *
   * @param isi
   * the first itemset to compare
   *
   * @param isj
   * the itemset to compare to the first
   *
   * @return true if it is ok to prune the itemsets that
   * would be generated with isi and larger isj,
   * otherwise false
   */

  private boolean pruneCandidateGeneration ( ARMinerItemset isi, 
					     ARMinerItemset isj )
  {
    // if this feature is off just return false
    switch ( usePruneCandidateGeneration ) {
    case NO:
      // report that it is turned off
      switch ( debug ) {
      case 1:
	System.out.println ( "ARMinerApriori.pruneCandidateGeneration: " +
			     "feature is off. " +
			     "Should be off for testing only." );
      } // end switch debug
      
      return ( false );
      
    } // end switch use prune candidate generation

    // we will only look through the itemsets once, 
    // starting with the first item
    int i = 0;
    int j = 0;
    
    // look at every item except the last
    for ( ; i < isi.size() - 1; i++ )
    {
      // if itemset isi has a regular item (NOT an event) as the first
      // item and itemset isj does not have the same item
      // we can return true (in other words prune this branch of candidate
      // generation) just like we didn't have events.
      
      if ( isi.eventName ( isi.getItem ( i ) ) == null )
      {
	// look at every item except the last
	for ( ; j < isj.size() - 1; j++ )
	{
	  if ( isi.getItem ( i ) == isj.getItem ( j ) )
	  {
	    break;

	  } // end if we can continue with the next i item
	  else
	  {
	    // if the item's integer value
	    // is greater than item i we can return true to prune
	    if ( isj.getItem ( j ) > isi.getItem ( i ) )
	    {
	      return ( true );
	      
	    } // end if item j is larger than i 
	    
	  } // end else first items in i and j don't match
	  
	} // end for each item in isj
	
      } // end if item is NOT an event
      
    } // end for each item in isi
    
    // well, I guess we found them all and we shouldn't prune
    return ( false );

  } // END private boolean pruneCandidateGeneration ( ARMinerItemset isi, 
  //				      ARMinerItemset is2 )

//********************************************************//

  /** 
   * Returns the number of event items of the same kind
   * allowed in an itemset.
   *
   * @return the maximum number of event items
   */

  public int getMaxEvents()
  {
    return ( maxEvents );
  } // END public int getMaxEvents()
 
//********************************************************//

  /** 
   * Sets the number of event items of the same kind
   * allowed in an itemset.
   *
   * @param num
   * the maximum number of event items
   */

  public void setMaxEvents ( int num )
  {
    maxEvents =num;
  } // END public void setMaxEvents()
 
//********************************************************//

  /**
   * Sets the logging mechanism. Specifying null turns
   * logging off.
   *
   * @param log
   * the logging mechanism, null turns logging off
   *
   * @return true if logging is on, false if logging off
   */

  public boolean setLogger ( Logger log )
  {
    logger = log;

    return ( logger != null );

  } // END public boolean setLogger ( Logger log )

//********************************************************//

} // END public class ARMinerApriori implements LargeItemsetsFinder
