/*
ARMiner - Association Rules Miner
Copyright (C) 2000  UMass/Boston - Computer Science Department

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The ARMiner Server was written by Dana Cristofor and Laurentiu
Cristofor.

The ARMiner Client was written by Abdelmajid Karatihy, Xiaoyong Kuang,
and Lung-Tsung Li.

The ARMiner package is currently maintained by Laurentiu Cristofor
(laur@cs.umb.edu).
*/

package wpi.associations.arminerSequence;

import java.util.*;
import java.io.IOException;
import java.io.EOFException;

/*
  Maintainance log started on November 30th, 2000

  Nov. 30th, 2000   - fixed rule generation procedure to use
                      all frequent itemsets, not only the
                      maximal ones

  Feb. 14th, 2002   - changed constraints to work with ARMinerItemsets
                      so that one of a group of items would be
                      required -zss

		      Note: this class contains pieces which are not 
		      used by the WEKA ARMiner implementation.  
		      I have tried to alter it as little as possible. -zss

  Apr. 30th, 2002   - added support for time sequence attributes
                    - added debug info
                    - cleaned up formatting
		    - added comments -kap

  February 9, 2003  - added removal of duplicate rules -kap

  February 24, 2003 - made consistent the convention of using a lower case
		      letter to start a method name. -kap

  August 13, 2003   - added numInstances data member, and get and set 
                      methods -kap
		    - now sets the num instances in all rules produced -kap

  December 2, 2003  - moved buildItemset method to class ARMinerItemset. -kap

  December 11, 2003 - modified buildRule method to not calculate confidence
                      for rules whose antecedent and consequent contain event
		      items. The calling method is responsible for calculating
		      the confidence. -kap
*/

/*
   This file is a part of the ARMiner project.
   
   (P)1999-2000 by ARMiner Server Team:

   Dana Cristofor
   Laurentiu Cristofor

   modified by Chris Shoemaker in March 2001
   modified by Zack Stoecker-Sylvia in February 2002
   modified by Keith A. Pray in April 2002
*/

/**
   AprioriRules.java<P>

   This class implements the Apriori algorithm 
   for finding association rules.

   (see "Fast Algorithms for Mining Association Rules"
   by Rakesh Agrawal and Ramakrishnan Srikant
   from IBM Almaden Research Center 1994)

@author Keith A. Pray (kap@wpi.edu)
*/

public class AprioriRules implements AssociationsFinder 
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  private SET supports;
  private Vector rules;
  
  // changed naming convention... underscores are icky -kap

  private float minSupport;
  private float minConfidence;

  private ARMinerItemset is_in_antecedent;
  private ARMinerItemset is_in_consequent;
  private ARMinerItemset is_ignored;

  // adding min and max for both -kap
  private int minAntecedent;
  private int maxAntecedent;
  private int minConsequent;
  private int maxConsequent;

  // adding members for required antecedents and consequents -kap

  /** Required antecedents for rule generated */
  private Vector requiredAntecedents;

  /** Required consequents for rule generated */
  private Vector requiredConsequents;

  /** Number of instances in the data set. Used for computing statistical
      measures */
  private int numInstances;

  // ----- Stuff needed for time sequence attribute support. ----- //
  
  /** Used to look up actual value of an item using its 
      integer representation. */
  private Hashtable numberHash;

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: and output from methods
   * 3: and all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Sets the number hash to use for looking up values
   * in this itemset
   *
   * @param h
   * the number hash
   */

  public void setNumberHash ( Hashtable h )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "AprioriRules: Set number hash: size = " + 
			   h.size() );
    } // end switch debug -kap

    numberHash = h;

  } // END public void setNumberHash ( Hashtable h )

//********************************************************//

  /**
   * Returns the number hash used for looking up values
   * in this itemset.
   *
   * @return the number hash
   */

  public Hashtable getNumberHash()
  {
    return ( numberHash );

  } // END public Hashtable getNumberHash()

//********************************************************//

  /**
   * Sets the number of instances contained in the data set
   * from which the frequent itemset were mined.
   *
   * @param n
   * the number of instances or rows
   */

  public void setNumInstances ( int n )
  {
    numInstances = n;

  } // END public void setNumInstances ( int n )

//********************************************************//

  /**
   * Returns the number of instances contained in the data set
   * from which the frequent itemset were mined.
   *
   * @return the number of instances or rows
   */

  public int getNumInstances()
  {
    return ( numInstances );

  } // END public int getNumInstances()

//********************************************************//

  /**
   * This method stores all frequent itemsets that have
   * support greater than the minimum support in a SET for
   * more efficient access times.
   */

  private void initializeSupports ( DBCacheReader cacheReader ) 
  {
    // create new SET
    supports = new SET();
    
    // we must pass the number hash
    // to the new SET to support time sequence
    // attributes -kap
    supports.setNumberHash ( numberHash );
    
    try 
    {
      ARMinerItemset is;
      while ( true ) 
      {
	// get item from cache
	is = cacheReader.getNextItemset();

	// if item has support greater than the minimum support
	// required then we add it to the SET
	if ( is.getSupport() >= minSupport ) 
	{
	  switch ( debug ) {
	  case 3: case 2: case 1: //case 0:
	    System.out.println ( "AprioriRules.initializeSupports: " +
				 "adding to supports: itemset = " +
				 is + " " );
	  } // end switch debug -kap

	  supports.insert ( is );
	}

      } // end while

    } // end try
    catch ( EOFException e ) 
    {
      // do nothing, we just reached the EOF
    } 
    catch ( IOException e ) 
    {
      System.err.println ( "Error scanning cache!!!\n" + e );
    } 
    catch ( ClassNotFoundException e ) 
    {
      System.err.println ( "Error scanning cache!!!\n" + e );
    }

  } // END private void initializeSupports ( DBCacheReader cacheReader ) 

//********************************************************//

  /**
   * Find association rules in a database, given the set of
   * frequent itemsets.
   *
   * @param cacheReader  
   * the object used to read from the cache
   *
   * @param min_Support  
   * the minimum support
   *
   * @param min_Confidence  
   * the minimum confidence
   *
   * @return a Vector containing all association rules found
   */

  public Vector findAssociations ( DBCacheReader cacheReader,
				   float min_Support,
				   float min_Confidence ) 
  {
    minSupport = min_Support;
    minConfidence = min_Confidence;

    // create the vector where we'll put the rules
    rules = new Vector();
    
    // read from cache supports of frequent itemsets
    initializeSupports ( cacheReader );

    // get the frequent itemsets
    // it looks like this is wasteful, why not read them from the cache?
    // why not have initialize supports save the itemsets in a vector
    // at the same time it feeds each one to 
    // supports (a SET object, prefix tree)? -kap
    Vector frequent = supports.getItemsets();

    // generate rules from each frequent itemset
    for ( int i = 0; i < frequent.size(); i++ ) 
    {
      // get a frequent itemset
      ARMinerItemset is_frequent = (ARMinerItemset)frequent.get ( i );

      // skip it if it's too small
      if ( is_frequent.size() <= 1 )
      {
	continue;
      }

      // get all possible 1 item consequents
      Vector consequents = new Vector ( is_frequent.size() );

      for ( int k = 0; k < is_frequent.size(); k++ ) 
      {
	int item = is_frequent.getItem ( k );
	ARMinerItemset is_consequent = new ARMinerItemset ( 1 );

	// we must pass the number hash
	// to the new itemset to support time sequence
	// attributes -kap
	is_consequent.setNumberHash ( numberHash );

	// at least use the spiffy method to use another itemset to get
	// name and values from, to save time -kap
	//	is_consequent.addItem ( item );
	is_consequent.addItem ( item, is_frequent );
	
	// is_consequent now contains a possible consequent
	// verify now that the rule having this consequent
	// satisfies our requirements
	
	ARMinerItemset is_antecedent = is_frequent.subtract ( is_consequent );

	is_antecedent.setSupport ( (float)0.000001 );
	is_consequent.setSupport ( (float)0.000001 );

	// retrieve the supports from the SET supports -kap
	try 
	{
	  is_antecedent.setSupport ( supports.getSupport ( is_antecedent ) );
	  is_consequent.setSupport ( supports.getSupport ( is_consequent ) );
	} 
	catch ( SETException e ) 
	{
	  System.err.println ( "AprioriRules.findAssociations: " +
			       "Error getting support from SET.\n" + e );
	} // end catch set exception
	 
	float confidence = 
	  is_frequent.getSupport() / is_antecedent.getSupport();

	if ( confidence >= minConfidence ) 
	{
	  consequents.add ( is_consequent );
	  
	  // we add the rule to our collection if it satisfies
	  // our conditions

	  // have to pass the original itemset in too to support
	  // time sequence stuff -kap
	  rules.add ( new AssociationRule ( is_antecedent, is_consequent,
					    is_frequent,  // the original
					    is_frequent.getSupport(), 
					    confidence ) );
	  
	  // set the number of instances so neat statistical things
	  // like chi-square can be calculated
	  ((AssociationRule)rules.lastElement()).
	    setNumInstances ( numInstances );

	} // end if confidence above min confidence

      } // end of for ( k )

      // call the ap_genrules procedure for generating all rules
      // out of this frequent itemset
      ap_genrules ( is_frequent, consequents );

    } // end of for ( i )
    
    return ( rules );
    
  } // END public Vector findAssociations ( DBCacheReader cacheReader,
  //				   float minSupport,
  //				   float minConfidence) 

//********************************************************//

  /**
   * This is the ap-genrules procedure that generates
   * rules out of a frequent itemset.
   */

  private void ap_genrules ( ARMinerItemset is_frequent, 
			     Vector consequents ) 
  {
    if ( consequents.size() == 0 )
    {
      return;
    }
    
    // the size of frequent must be bigger than the size of the itemsets
    // in consequents by at least 2, in order to be able to generate
    // a rule in this call
    if ( is_frequent.size() > 
	 ((ARMinerItemset)(consequents.get ( 0 ) )).size() + 1 ) 
    {
      // get a list of ( k + 1 )-itemsets from consequents. WHY? -kap
      Vector new_consequents = apriori_gen ( consequents );
      AssociationRule ar;

      for ( int i = 0; i < new_consequents.size(); i++ ) 
      {
	ARMinerItemset is_consequent = 
	  (ARMinerItemset)new_consequents.get ( i );

	ARMinerItemset is_antecedent = is_frequent.subtract ( is_consequent );

	switch ( debug ) {
	case 3: case 2: case 1: //case 0:
	  System.out.println ( "AprioriRules.ap_genrules: " +
			       "is_frequent has support = " +
			       is_frequent.getSupport() + " " +
			       "is_consequent has support = " +
			       is_consequent.getSupport() + " " +
			       "is_antecendent has support = " +
			       is_antecedent.getSupport() );
	} // end switch debug -kap

	is_antecedent.setSupport ( (float)0.00001 );
	is_consequent.setSupport ( (float)0.00001 );

	try 
	{
	  is_antecedent.setSupport ( supports.getSupport ( is_antecedent ) );
	  is_consequent.setSupport ( supports.getSupport ( is_consequent ) );
	} 
	catch ( SETException e ) 
	{
	  System.err.println ( "AprioriRules.ap_genrules_constraint: " +
			       "Error getting support from SET.\n" + e );
        } // end catch set exception

	float confidence = 
	  is_frequent.getSupport() / is_antecedent.getSupport();
	    
	// if the rule satisfies our requirements we add it
	// to our collection
	if ( confidence >= minConfidence )
	{
	  // have to pass the original itemset in too to support
	  // time sequence stuff -kap
	  rules.add ( new AssociationRule ( is_antecedent, is_consequent,
					    is_frequent,  // the original
					    is_frequent.getSupport(), 
					    confidence ) );

	  // set the number of instances so neat statistical things
	  // like chi-square can be calculated
	  ((AssociationRule)rules.lastElement()).
	    setNumInstances ( numInstances );

	} // end if confidence above threshold

	// otherwise we remove the consequent from the collection
	// and we update the index such that we don't skip a consequent
	else
	{
	  new_consequents.remove ( i-- );
	}

      } // end for 

      ap_genrules( is_frequent, new_consequents );
    }

  } // END private void ap_genrules ( ARMinerItemset is_frequent, 
  //			     Vector consequents ) 

//********************************************************//

  /**
   * This is the apriori_gen procedure that generates
   * starting from a k-itemset collection a new collection
   * of (k+1)-itemsets.
   */
  
  private Vector apriori_gen ( Vector itemsets ) 
  {
    if ( itemsets.size() == 0 )
    {
      return ( new Vector ( 0 ) );

    } // end if there are no itemsets

    // create a hashtree so that we can check more efficiently the
    // number of subsets
    // this may not really be necessary when generating rules since
    // itemsets will probably be a small collection, but just in case

    // let's only do this if there are no event items in the data set

/*    HashTree ht_itemsets = new HashTree ( itemsets );

    for ( int i = 0; i < itemsets.size(); i++ )
    {
      ht_itemsets.add ( i );
    }

   ht_itemsets.prepareForDescent();
*/
    Vector result = new Vector();
    ARMinerItemset is_i, is_j;

    for ( int i = 0; i < itemsets.size() - 1; i++ )
    {
      for ( int j = i + 1; j < itemsets.size(); j++ ) 
      {
	is_i = (ARMinerItemset)itemsets.get ( i );
	is_j = (ARMinerItemset)itemsets.get ( j );
	
	// if we cannot combine element i with j then we shouldn't
	// waste time for bigger j's. This is because we keep the
	// collections ordered, an important detail in this implementation

	// to support time sequence stuff can combine with now returns
	// a hastbale mapping event labels. If it is null then the
	// itemsets cannot be combined. This hashtable is now required
	// by the ARMinerItemset.combineWith method.
	Hashtable eventLabelMap = is_i.canCombineWith ( is_j );

	if ( eventLabelMap == null )
	{
//	  break;
	}
	else 
	{
	  // to support time sequences combineWith has been changed to
	  // return a list of itemsets and take a hashtable mapping
	  // event labels -kap
	  //	      ARMinerItemset is = is_i.combineWith(is_j);
	  
	  Enumeration e = is_i.combineWith ( is_j, eventLabelMap ).elements();
	  ARMinerItemset is;
	  
	  // so now do the same thing for each itemset in the list -kap
	  while ( e.hasMoreElements() )
	  {
	    is = (ARMinerItemset)e.nextElement();
	    
	    // a real k-itemset has k (k-1)-subsets
	    // so we test that this holds before adding to result
//	    if ( ht_itemsets.countSubsets ( is ) == is.size() )
//	    {
	      result.add ( is );
//	    }

	  } // end while e has more -kap
	  
	} // end else
	
      } // end for

    } // end for 

    return ( result );

  } // END private Vector apriori_gen ( Vector itemsets ) 

//********************************************************//

  /**
   * Find association rules in a database, given the set of
   * frequent itemsets and a set of restrictions.
   *
   * @param cacheReader  
   * the object used to read from the cache
   *
   * @param min_Support   the minimum support
   *
   * @param min_Confidence   the minimum confidence
   *
   * @param inAntecedent   the items that must appear in the antecedent 
   * of each rule, if null then this constraint is ignored 
   *
   * @param inConsequent   the items that must appear in the consequent
   * of each rule, if null then this constraint is ignored 
   *
   * @param ignored  
   * the items that should be ignored,
   * if null then this constraint is ignored 
   *
   * @param max_Antecedent  
   * the maximum number of items that can appear
   * in the antecedent of each rule, if 0 then this constraint is ignored 
   *
   * @param min_Consequent  
   * the minimum number of items that should appear
   * in the consequent of each rule, if 0 then this constraint is ignored 
   *
   * @return a Vector containing all association rules found
   */

  public Vector findAssociations ( DBCacheReader cacheReader, 
				   float min_Support,
				   float min_Confidence,
				   ARMinerItemset inAntecedent,
				   ARMinerItemset inConsequent,
				   ARMinerItemset ignored,
				   int max_Antecedent, 
				   int min_Consequent ) 
  {
    minSupport = minSupport;
    minConfidence = minConfidence;
    
    is_in_antecedent = inAntecedent;
    is_in_consequent = inConsequent;
    is_ignored = ignored;
    maxAntecedent = max_Antecedent;
    minConsequent = min_Consequent;
    
    // create the vector where we'll put the rules
    rules = new Vector();
    
    // read from cache supports of frequent itemsets
    initializeSupports ( cacheReader );
    
    // get the frequent itemsets
    Vector frequent = supports.getItemsets();

    if ( frequent.size() == 0 )
    {
      return ( rules );
    }
    
    // if we need to ignore some items
    if ( ignored != null ) 
    {
      // remove all frequent itemsets that contain
      // items to be ignored; their subsets that do
      // not contain those items will remain
      for ( int i = 0; i < frequent.size(); i++ ) 
      {
	ARMinerItemset is = (ARMinerItemset)frequent.get ( i );

	if ( is.doesIntersect ( ignored ) ) 
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  frequent.set ( i, frequent.lastElement() );
	  frequent.remove ( frequent.size() - 1 );
	  i--;
	}
      }

      if ( frequent.size() == 0 )
      {
	return ( rules );
      }
    }
    
    // if we need to have some items in the antecedent or consequent
    if ( inAntecedent != null || inConsequent != null ) 
    {
      // remove frequent itemsets that don't have the
      // required items
      for ( int i = 0; i < frequent.size(); i++ ) 
      {
	ARMinerItemset is = (ARMinerItemset)frequent.get ( i );

	if ( inAntecedent != null && ! inAntecedent.isIncludedIn ( is ) )
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  frequent.set ( i, frequent.lastElement() );
	  frequent.remove ( frequent.size() - 1 );
	  i--;

	  //CAS: 
	  //} else if (inConsequent != null && !inConsequent.isIncludedIn(is)) {
	} 
	else 
	  if ( inConsequent != null && ! inConsequent.doesIntersect ( is ) ) 
	  {
	    //} else if (inConsequent != null && !is.isIncludedIn(inConsequent)) {
	    // replace this element with last, delete last,
	    // and don't advance index
	    frequent.set ( i, frequent.lastElement() );
	    frequent.remove ( frequent.size() - 1 );
	    i--;
	  }
      } // end of for(i) - removing itemsets that don't have the items

      if ( frequent.size() == 0 )
	return rules;

    } // end of if inAntecedent != null or inConsequent != null

    // generate rules from each frequent itemset
    for ( int i = 0; i < frequent.size(); i++ ) 
    {
      // get a frequent itemset
      ARMinerItemset is_frequent = (ARMinerItemset)frequent.get ( i );
      
      // skip it if it's too small
      if ( is_frequent.size() <= 1 ||
	   is_frequent.size() <= minConsequent )
	continue;

      // get all possible 1 item consequents
      Vector consequents = new Vector ( is_frequent.size() );

      for ( int k = 0; k < is_frequent.size(); k++ ) 
      {
	int item = is_frequent.getItem ( k );
	ARMinerItemset is_consequent = new ARMinerItemset ( 1 );
	
	// we must pass the number hash
	// to the new itemset to support time sequence
	// attributes -kap
	is_consequent.setNumberHash ( numberHash );
	
	is_consequent.addItem ( item );
	
	// is_consequent now contains a possible consequent
	// verify now that the rule having this consequent
	// satisfies our requirements
	
	ARMinerItemset is_antecedent = 
	  is_frequent.subtract ( is_consequent );

	is_antecedent.setSupport ( (float)0.00001 );
	is_consequent.setSupport ( (float)0.00001 );

	try 
	{
	  is_antecedent.setSupport ( supports.getSupport ( is_antecedent ) );
	  is_consequent.setSupport ( supports.getSupport ( is_consequent ) );
	} 
	catch ( SETException e ) 
	{
	  System.err.println ( "AprioriRules.ap_genrules_constraint: " +
			       "Error getting support from SET.\n" + e );
	} // end catch set exception
	
	float confidence = 
	  is_frequent.getSupport() / is_antecedent.getSupport();
	
	if ( confidence >= minConfidence ) 
	{
	  consequents.add ( is_consequent );
	  
	  // check whether it also satisfies our constraints
	  boolean approved = true;
	  
	  if ( approved && is_in_antecedent != null
	       && ! is_in_antecedent.isIncludedIn ( is_antecedent ) )
	    approved = false;
	  
          // *** CAS: changed consequent constraint from requiring ALL
          // items to be present in the consequent to requiring AT LEAST one
          // item to be present in the consequent
          // #### Updated change to require that consequent is a SUBSET of
          // consequent constraint.
	  if ( approved && is_in_consequent != null
	       //  && !is_in_consequent.isIncludedIn(is_consequent))
	       //  && !is_in_consequent.doesIntersect(is_consequent))
	       && ! is_consequent.isIncludedIn ( is_in_consequent ) )
	    approved = false;

	  if ( approved && maxAntecedent > 0
	       && is_antecedent.size() > maxAntecedent )
	    approved = false;

	  if ( approved && minConsequent > 0 
	       && is_consequent.size() < minConsequent )
	    approved = false;

	  // if the rule satisifes all requirements then
	  // we add it to the rules collection
	  if ( approved )
	  {
	    // have to pass the original itemset in too to support
	    // time sequence stuff -kap
	    rules.add ( new AssociationRule ( is_antecedent, 
					      is_consequent,
					      is_frequent,  // the original
					      is_frequent.getSupport(), 
					      confidence ) );

	    // set the number of instances so neat statistical things
	    // like chi-square can be calculated
	    ((AssociationRule)rules.lastElement()).
	      setNumInstances ( numInstances );

	  } // end if approved

	} // end of if confidence >= min_confidence
      } // end of for(k)

      // call the ap-genrules procedure for generating all rules
      // out of this frequent itemset
      ap_genrules_constraint ( is_frequent, consequents );
    } // end of for(i)

    return ( rules );

  } // END public Vector findAssociations ( DBCacheReader cacheReader, 
  //				   float minSupport,
  //				   float minConfidence,
  //				   ARMinerItemset inAntecedent,
  //				   ARMinerItemset inConsequent,
  //				   ARMinerItemset ignored,
  //				   int maxAntecedent, 
  //				   int minConsequent ) 

//********************************************************//

  /**
   * ------------------- WEKA VERSION ------------------
   * 
   * Find association rules in a database, given the set of
   * frequent itemsets and a set of restrictions.
   *
   * @param cacheReader	
   * the object used to read from the cache
   *
   * @param min_Support
   * the minimum support
   *
   * @param min_Confidence     
   * the minimum confidence
   *
   * @param antecedentAttributes	
   * a vector of ARMinerItemsets that each contains only items in a single
   * attribute used to restrict which attributes must appear in the
   * antecedent, ignored if null
   *
   * @param consequentAttributes	
   * a vector of ARMinerItemsets that each contains only items in a single
   * attribute used to restrict which attributes must appear in the
   * consequent, ignored if null
   *
   * @param max_Antecedent
   * the maximum number of items that can appear in the antecedent of each
   * rule, if 0 then this constraint is ignored 
   *
   * @param min_Consequent   
   * the minimum number of items that should appear in the consequent of
   * each rule, if 0 then this constraint is ignored 
   *
   * @return
   * a Vector containing all association rules found
   */

  public Vector findAssociations ( DBCacheReader cacheReader, 
				   float min_Support,
				   float min_Confidence,
				   Vector antecedentAttributes,
				   Vector consequentAttributes,
				   int max_Antecedent, 
				   int min_Consequent )
  {
    minSupport = min_Support;
    minConfidence = min_Confidence;
    
    maxAntecedent = max_Antecedent;
    minConsequent = min_Consequent;

    // create the vector where we'll put the rules
    rules = new Vector();
    
    // read from cache supports of frequent itemsets
    initializeSupports ( cacheReader );

    // get the frequent itemsets
    //    Vector frequent = supports.getItemsets();
    Vector frequent = supports.getItemsets();
    
    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "AprioriRules.findAssociations: " +
			   "(weka version): " +
			   "supports=\n" + supports );
    case 1: //case 0:
      System.out.println ( "AprioriRules.findAssociations: " +
			   "(weka version): " +
			   "frequent itemsets=\n" + frequent );
    } // end switch debug

    if ( frequent.size() == 0 )
    {
      return ( rules );
    }

    // Let's speed up the removing of frequent itemsets not
    // fitting our criteria of size and required antecedents
    // and consequents -kap
    
    // things to check, 
    // 0 = check neither
    // 1 = check antecedent
    // 2 = check consequent
    // 3 = check both
    int check = 0;

    if ( antecedentAttributes != null && antecedentAttributes.size() != 0)
    {
      requiredAntecedents = new Vector ( antecedentAttributes );
      check = 1;
    } // end if we don't have to check for antecedents

    if ( consequentAttributes != null && consequentAttributes.size() != 0)
    {
      requiredConsequents = new Vector ( consequentAttributes );
      switch ( check )
      {
      case 0:
	check = 2;
	break;
      case 1:
	check = 3;
	break;
      } //end switch default

    } // end if we don't have to check for antecedents

    // if we need to have some items in the antecedent or consequent
    //    if ( antecedentAttributes != null || consequentAttributes != null ) 
    //    {

      // remove frequent itemsets that don't have the required items
      // ... and are too small -kap
      for ( int i = 0; i < frequent.size(); i++ ) 
      {
	ARMinerItemset is = (ARMinerItemset)frequent.get ( i );
	
	// might as well weed out itemset that are too small here too -kap
	if ( is.size() < 2 || is.size() <= minConsequent )
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  frequent.set ( i, frequent.lastElement() );
	  frequent.remove ( frequent.size() - 1 );
	  i--;
	} // end if itemset too small
	else
	{
	  switch ( check )
	  {
	  case 1: // check antecedent
	    if ( ! is.hasAttributes ( requiredAntecedents ) )
	    {
	      // replace this element with last, delete last,
	      // and don't advance index
	      frequent.set ( i, frequent.lastElement() );
	      frequent.remove ( frequent.size() - 1 );
	      i--;
	    }
	    break;
	  case 2:
	    if ( ! is.hasAttributes ( requiredConsequents ) )
	    {
	      // replace this element with last, delete last,
	      // and don't advance index
	      frequent.set ( i, frequent.lastElement() );
	      frequent.remove ( frequent.size() - 1 );
	      i--;
	      
	    } // end if itemset missing required consequent
	    break;
	  case 3:
	    if ( !is.hasAttributes ( requiredAntecedents ) )
	    {
	      // replace this element with last, delete last,
	      // and don't advance index
	      frequent.set ( i, frequent.lastElement() );
	      frequent.remove ( frequent.size() - 1 );
	      i--;
	    }
	    else
	    {
	      if ( !is.hasAttributes ( requiredConsequents ) )
	      {
		// replace this element with last, delete last,
		// and don't advance index
		frequent.set ( i, frequent.lastElement() );
		frequent.remove ( frequent.size() - 1 );
		i--;
		
	      } // end if itemset missing required consequent

	    } // end else itemset had required antecedents

	  } // end switch check
	    
	} // end else itemset large enough
	  
      } // end for ( i ) - removing itemsets that don't have required items
      
      if ( frequent.size() == 0 )
      {
	return ( rules );
	
      } // end if there are no frequent itemsets left
      
    //} // end if antecedentAttributes != null or consequentAttributes != null
    
    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "AprioriRules.findAssociations: " +
			   "(weka version): " +
			   "frequent itemsets after removing those not " +
			   "containing required antecedent and consequents= " +
			   frequent );
    } // end switch debug

    // let's do something different...
    // using the maximal frequent itemsets let's generate all the same rules
    
    // generate rules from each frequent itemset
    for ( int i = 0; i < frequent.size(); i++ ) 
    {
      // get a frequent itemset
      ARMinerItemset is_frequent = (ARMinerItemset)frequent.get ( i );

      // removed these itemset above instead... faster that way -kap
      // skip it if it's too small
      /* if ( is_frequent.size() <= 1 || is_frequent.size() <= minConsequent )
      {
	continue;
      }
      */

      // get all possible 1 item consequents
      Vector consequents = new Vector ( is_frequent.size() );

      for ( int k = 0; k < is_frequent.size(); k++ ) 
      {
	int item = is_frequent.getItem ( k );
	ARMinerItemset is_consequent = new ARMinerItemset ( 1 );
	
	// We must pass the number hash to the new itemset to
	// support time sequence attributes. -kap
	is_consequent.setNumberHash ( numberHash );

	switch ( debug ) {
	case 3: //case 0:
	  System.out.println ( "AprioriRules.findAssociations: " +
			       "(weka version): " +
			       "adding item " + item + " with value= " +
			       numberHash.get ( new Integer ( item ) ) );
	} // end switch debug

	// Use the nifty method to add an item that gets the values
	// from nother itemset so we don't have to parse the value -kap
	//	is_consequent.addItem ( item );

	is_consequent.addItem ( item, is_frequent );

	// is_consequent now contains a possible consequent
	// verify now that the rule having this consequent
	// satisfies our requirements
	
	ARMinerItemset is_antecedent = is_frequent.subtract ( is_consequent );

	is_antecedent.setSupport ( (float)0.00001 );
	is_consequent.setSupport ( (float)0.00001 );

	try 
	{
	  is_antecedent.setSupport ( supports.getSupport ( is_antecedent ) );
	  is_consequent.setSupport ( supports.getSupport ( is_consequent ) );
	} 
	catch ( SETException e ) 
	{
	  System.err.println ( "AprioriRules.ap_genrules_constraint: " +
			       "Error getting support from SET.\n" + e );
	} // end catch set exception

	float confidence = 
	  is_frequent.getSupport() / is_antecedent.getSupport();

	switch ( debug ) {
	case 3: case 2: case 1: //case 0:
	  System.out.println ( "AprioriRules.findAssociations " +
			       "(weka version): " +
			       "(is_frequent support = " +
			       is_frequent.getSupport() + " " +
			       ")  (is_antecedent support = " +
			       is_antecedent.getSupport() + ")" );
	} // end switch debug -kap
	
	if ( confidence >= minConfidence ) 
	{
	  consequents.add ( is_consequent );
	  
	  // check whether it also satisfies our constraints
	  boolean approved = true;
	  
	  if ( approved && maxAntecedent > 0
	       && is_antecedent.size() > maxAntecedent )
	    approved = false;
	  
	  if ( approved && minConsequent > 0 
	       && is_consequent.size() < minConsequent )
	    approved = false;
	  
	  if ( approved && requiredAntecedents != null
	       && !is_antecedent.hasAttributes ( requiredAntecedents ) )
	    approved = false;
	  
	  if ( approved && requiredConsequents != null
	       && !is_consequent.hasAttributes ( requiredConsequents ) )
	    approved = false;
	  
	  // if the rule satisifes all requirements then
	  // we add it to the rules collection
	  if ( approved )
	  {
	    // have to pass the original itemset in too to support
	    // time sequence stuff -kap
	    rules.add ( new AssociationRule ( is_antecedent, 
					      is_consequent,
					      is_frequent,  // the original
					      is_frequent.getSupport(), 
					      confidence ) );

	    // set the number of instances so neat statistical things
	    // like chi-square can be calculated
	    ((AssociationRule)rules.lastElement()).
	      setNumInstances ( numInstances );

	  } // end if rule approved

	} // end of if confidence >= min_confidence

      } // end of for ( k )
      
      // call the ap-genrules procedure for generating all rules
      // out of this frequent itemset that checks for required
      // antecedents and consequents -kap
      ap_genrules_constraint ( is_frequent, consequents, 
			       requiredAntecedents, requiredConsequents );
    } // end of for ( i )

    return ( rules );

  } // END public Vector findAssociations ( DBCacheReader cacheReader, 
  //				   float minSupport,
  //				   float minConfidence,
  //				   Vector antecedentAttributes,
  //				   Vector consequentAttributes,
  //				   int maxAntecedent, 
  //				   int minConsequent )

//********************************************************//

  // this takes the place of the previous weka version -kap

  /**
   * -------------- ANOTHER VERSION FOR WEKA -------------
   *
   * Find association rules in a database, given the set of
   * frequent itemsets and a set of restrictions.
   *
   * @param cacheReader	
   * the object used to read from the cache
   *
   * @param minSupport
   * the minimum support
   *
   * @param minConfidence     
   * the minimum confidence
   *
   * @param antecedentAttributes	
   * a vector of ARMinerItemsets that each contains only items in a single
   * attribute used to restrict which attributes must appear in the
   * antecedent, ignored if null
   *
   * @param consequentAttributes	
   * a vector of ARMinerItemsets that each contains only items in a single
   * attribute used to restrict which attributes must appear in the
   * consequent, ignored if null
   *
   * @param minAntecedent
   * the minimum number of items that can appear in the antecedent of each
   * rule, if 0 then this constraint is ignored 
   *
   * @param maxAntecedent
   * the maximum number of items that can appear in the antecedent of each
   * rule, if 0 then this constraint is ignored 
   *
   * @param minConsequent   
   * the minimum number of items that should appear in the consequent of
   * each rule, if 0 then this constraint is ignored 
   *
   * @param maxConsequent   
   * the minimum number of items that should appear in the consequent of
   * each rule, if 0 then this constraint is ignored 
   *
   * @return
   * a Vector containing all association rules found
   */

  public Vector findAssociations ( DBCacheReader cacheReader, 
				   float min_Support,
				   float min_Confidence,
				   Vector antecedentAttributes,
				   Vector consequentAttributes,
				   int min_Antecedent, 
				   int max_Antecedent, 
				   int min_Consequent,
				   int max_Consequent )
  {
    // save all the criteria
    minSupport = min_Support;
    minConfidence = min_Confidence;
    
    minAntecedent = min_Antecedent;
    maxAntecedent = max_Antecedent;
    minConsequent = min_Consequent;
    maxConsequent = max_Consequent;

    // except the required attributes... for now
    requiredAntecedents = null;
    requiredConsequents = null;

    // create the vector where we'll put the rules
    rules = new Vector();
    
    // read from cache supports of frequent itemsets
    initializeSupports ( cacheReader );

    // get the maximal frequent itemsets
    Vector itemsets = supports.getMaximalItemsets();
    
    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "AprioriRules.findAssociations: " +
			   "(new weka version): " +
			   "supports=\n" + supports );
    case 1: //case 0:
      System.out.println ( "AprioriRules.findAssociations: " +
			   "(new weka version): " +
			   "frequent itemsets=\n" + itemsets );
    } // end switch debug

    if ( itemsets.size() == 0 )
    {
      return ( rules );
    }
    
    // Let's remove itemsets not fitting our criteria of size and
    //  required antecedents and consequents
    
    // things to check, 
    // 0 = check neither
    // 1 = check antecedent
    // 2 = check consequent
    // 3 = check both
    int check = 0;

    if ( antecedentAttributes != null && antecedentAttributes.size() != 0)
    {
      requiredAntecedents = new Vector ( antecedentAttributes );
      check = 1;
    } // end if we don't have to check for antecedents

    if ( consequentAttributes != null && consequentAttributes.size() != 0)
    {
      requiredConsequents = new Vector ( consequentAttributes );
      switch ( check )
      {
      case 0:
	check = 2;
	break;
      case 1:
	check = 3;
	break;
      } //end switch default
      
    } // end if we don't have to check for antecedents
    
    ARMinerItemset is;

    // remove frequent itemsets that don't have the required items
    // ... and are too small -kap
    for ( int i = 0; i < itemsets.size(); i++ ) 
    {
      is = (ARMinerItemset)itemsets.get ( i );
      
      // might as well weed out itemset that are too small here too -kap
      // added + minAntecedent and changed to < from <= -kap 2003-03-05
      if ( is.size() < 2 || is.size() < minConsequent + minAntecedent )
      {
	// replace this element with last, delete last,
	// and don't advance index
	itemsets.set ( i, itemsets.lastElement() );
	itemsets.remove ( itemsets.size() - 1 );
	i--;

	// skip the rest of the loop
	continue;
      } // end if itemset too small

      switch ( check )
      {
      case 1: // check antecedent
	if ( ! is.hasAttributes ( requiredAntecedents ) )
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  itemsets.set ( i, itemsets.lastElement() );
	  itemsets.remove ( itemsets.size() - 1 );
	  i--;
	}
	break;
      case 2:
	if ( ! is.hasAttributes ( requiredConsequents ) )
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  itemsets.set ( i, itemsets.lastElement() );
	  itemsets.remove ( itemsets.size() - 1 );
	  i--;
	  
	} // end if itemset missing required consequent
	break;
      case 3:
	if ( ! is.hasAttributes ( requiredAntecedents ) )
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  itemsets.set ( i, itemsets.lastElement() );
	  itemsets.remove ( itemsets.size() - 1 );
	  i--;
	}
	else
	{
	  if ( ! is.hasAttributes ( requiredConsequents ) )
	  {
	    // replace this element with last, delete last,
	    // and don't advance index
	    itemsets.set ( i, itemsets.lastElement() );
	    itemsets.remove ( itemsets.size() - 1 );
	    i--;
	    
	  } // end if itemset missing required consequent
	  
	} // end else itemset had required antecedents
	
      } // end switch check
      
    } // end for ( i ) - removing itemsets that don't have required items
    
    // do we have any itemsets left?
    if ( itemsets.size() == 0 )
    {
      return ( rules );
      
    } // end if there are no frequent itemsets left
    
    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "AprioriRules.findAssociations: " +
			   "(new weka version): " +
			   "frequent itemsets after removing those not " +
			   "containing required antecedent and consequents= " +
			   itemsets );
    } // end switch debug
    
    // using the maximal frequent itemsets let's generate all the rules

    // the maximal frequent itemsets are the set of itemsets that have the
    // most items in them and whose subsets constitute all the frequent
    // itemsets

    for ( int i = 0; i < itemsets.size(); i++ ) 
    {
      generateFromMaximal ( (ARMinerItemset)itemsets.get ( i ) );

    } // end for each itemset

    // have to make sure there are no duplicate rules
    // put it in a switch in case we want to report the
    // number of duplicate rules removed
    switch ( debug ) {
    case 1: //case 0:
      removeDuplicateRules();
      break;
    default:
      System.out.println ( "AprioriRules.findAssociations: " +
			   rules.size() +
			   " rules. " +
			   removeDuplicateRules() +
			   " duplicate rules removed. " +
			   rules.size() +
			   " rules left." );
    } // end switch debug

    return ( rules );

  } // END public Vector findAssociations ( DBCacheReader cacheReader, 
  //				   float minSupport,
  //				   float minConfidence,
  //				   Vector antecedentAttributes,
  //				   Vector consequentAttributes,
  //				   int minAntecedent, 
  //				   int maxAntecedent, 
  //				   int minConsequent,
  //				   int maxConsequent )

//********************************************************//  
  /**
   * This is the ap-genrules procedure that generates
   * rules out of a frequent itemset.
   */

  private void ap_genrules_constraint ( ARMinerItemset is_frequent, 
					Vector consequents, 
					Vector antecedentAttributes, 
					Vector consequentAttributes ) 
  {
    if ( consequents.size() == 0 )
    {
      return;
    }
    
    // the size of frequent must be bigger than the size of the itemsets
    // in consequents by at least 2, in order to be able to generate
    // a rule in this call
    if ( is_frequent.size() > 
	 ((ARMinerItemset)(consequents.get ( 0 ))).size() + 1 ) 
    {
      Vector new_consequents = apriori_gen ( consequents );

      // make sure each of the new consequents are included in is_frequent
      int i;
      
      for ( i = 0; i < new_consequents.size(); i++ )
      {
	if ( ! ((ARMinerItemset)new_consequents.get ( i )).
	     isIncludedIn ( is_frequent ) )
	{
	  // replace this element with last, delete last,
	  // and don't advance index
	  new_consequents.set ( i, new_consequents.lastElement() );
	  new_consequents.remove ( new_consequents.size() - 1 );
	  i--;
	  
	} // end if a consequent is not included

      } // end for each consequent

      AssociationRule ar;
      
      for ( i = 0; i < new_consequents.size(); i++ ) 
      {
	ARMinerItemset is_consequent = 
	  (ARMinerItemset)new_consequents.get ( i );

	ARMinerItemset is_antecedent = is_frequent.subtract ( is_consequent );

	is_antecedent.setSupport ( (float)0.00001 );
	is_consequent.setSupport ( (float)0.00001 );

	try 
	{
	  is_antecedent.setSupport ( supports.getSupport ( is_antecedent ) );
	  is_consequent.setSupport ( supports.getSupport ( is_consequent ) );
	} 
	catch ( SETException e ) 
	{
	  System.err.println ( "AprioriRules.ap_genrules_constraint: " +
			       "Error getting support from SET.\n" + e );
	} // end cacth set exception

	float confidence = 
	  is_frequent.getSupport() / is_antecedent.getSupport();
	
	// if the rule satisfies our confidence requirements
	if ( confidence >= minConfidence ) 
	{
	  // check whether it also satisfies our constraints
	  boolean approved = true;
	  
	  if ( approved && antecedentAttributes != null
	       && !is_antecedent.hasAttributes ( antecedentAttributes ) )
	  {
	    approved = false;

	  } // end if 
	  
          // *** CAS: changed consequent constraint from requiring ALL
          // items to be present in the consequent to requiring AT LEAST one
          // item to be present in the consequent
          // #### Updated change to require that consequent is a SUBSET of
          // consequent constraint.
	  if ( approved && consequentAttributes != null
	       //  && !is_in_consequent.isIncludedIn(is_consequent))
	       //  && !is_in_consequent.doesIntersect(is_consequent))
	       && !is_consequent.hasAttributes ( consequentAttributes ) )
	  {
	    approved = false;
	  }
	  
	  if ( approved && maxAntecedent > 0
	      && is_antecedent.size() > maxAntecedent )
	  {
	    approved = false;
	  }
	  
	  if ( approved && minConsequent > 0 
	      && is_consequent.size() < minConsequent )
	  {
	    approved = false;
	  }
	  
	  // if the rule satisifes all requirements then
	  // we add it to the rules collection
	  if ( approved )
	  {
	    // have to pass the original itemset in too to support
	    // time sequence stuff -kap
	    rules.add ( new AssociationRule ( is_antecedent, 
					      is_consequent,
					      is_frequent,  // the original
					      is_frequent.getSupport(), 
					      confidence ) );

	    // set the number of instances so neat statistical things
	    // like chi-square can be calculated
	    ((AssociationRule)rules.lastElement()).
	      setNumInstances ( numInstances );

	  } // end if approved

	} // end of if confidence >= min_confidence
	
	// otherwise we remove the consequent from the collection
	// and we update the index such that we don't skip a consequent
	else
	{
	  new_consequents.remove ( i-- );
	} // end else

      } // end of for(i)
      
      ap_genrules_constraint ( is_frequent, new_consequents );

    } // end of if

  } // END private void ap_genrules_constraint ( ARMinerItemset is_frequent, 
  //					Vector consequents, 
  //					Vector antecedentAttributes, 
  //					Vector consequentAttributes ) 

//********************************************************//
  
  /**
   * This is the ap-genrules procedure that generates rules out
   * of a frequent itemset.
   */

  private void ap_genrules_constraint ( ARMinerItemset is_frequent, 
					Vector consequents ) 
  {
    if ( consequents.size() == 0 )
    {
      return;
    } // end if there are no consequents

    // the size of frequent must be bigger than the size of the itemsets
    // in consequents by at least 2, in order to be able to generate
    // a rule in this call
    if ( is_frequent.size() > 
	 ((ARMinerItemset)(consequents.get ( 0 ))).size() + 1) 
    {
      Vector new_consequents = apriori_gen ( consequents );
      AssociationRule ar;
      
      for ( int i = 0; i < new_consequents.size(); i++ ) 
      {
	ARMinerItemset is_consequent = 
	  (ARMinerItemset)new_consequents.get ( i );

	ARMinerItemset is_antecedent = is_frequent.subtract ( is_consequent );

	is_antecedent.setSupport ( (float)0.00001 );
	is_consequent.setSupport ( (float)0.00001 );

	try 
	{
	  is_antecedent.setSupport ( supports.getSupport ( is_antecedent ) );
	  is_consequent.setSupport ( supports.getSupport ( is_consequent ) );
	} 
	catch ( SETException e ) 
	{
	  System.err.println ( "AprioriRules.ap_genrules_constraint: " +
			       "Error getting support from SET.\n" + e );
	}

	float confidence = 
	  is_frequent.getSupport() / is_antecedent.getSupport();
	
	// if the rule satisfies our confidence requirements
	if ( confidence >= minConfidence ) 
	{
	  // check whether it also satisfies our constraints
	  boolean approved = true;
	  
	  if ( approved && is_in_antecedent != null
	       && ! is_in_antecedent.isIncludedIn ( is_antecedent ) )
	  {
	    approved = false;
	  }
	  
          // *** CAS: changed consequent constraint from requiring ALL
          // items to be present in the consequent to requiring AT LEAST one
          // item to be present in the consequent
          // #### Updated change to require that consequent is a SUBSET of
          // consequent constraint.
	  if ( approved && is_in_consequent != null
	       //  && !is_in_consequent.isIncludedIn(is_consequent))
	       //  && !is_in_consequent.doesIntersect(is_consequent))
	      && ! is_consequent.isIncludedIn ( is_in_consequent ) )
	  {
	    approved = false;
	  }
	  
	  if ( approved && maxAntecedent > 0
	       && is_antecedent.size() > maxAntecedent )
	  {
	    approved = false;
	  }

	  if ( approved && minConsequent > 0 
	       && is_consequent.size() < minConsequent )
	  {
	    approved = false;
	  }

	  // if the rule satisifes all requirements then
	  // we add it to the rules collection
	  if ( approved )
	  {
	    // have to pass the original itemset in too to support
	    // time sequence stuff -kap
	    rules.add ( new AssociationRule ( is_antecedent, 
					      is_consequent,
					      is_frequent,  // the original
					      is_frequent.getSupport(), 
					      confidence ) );

	    // set the number of instances so neat statistical things
	    // like chi-square can be calculated
	    ((AssociationRule)rules.lastElement()).
	      setNumInstances ( numInstances );

	  } // end if approved

	} // end of if confidence >= min_confidence
	
	// otherwise we remove the consequent from the collection
	// and we update the index such that we don't skip a consequent
	else 
	{
	  new_consequents.remove ( i-- );
	} // end else

      } // end of for ( i )
      
      ap_genrules_constraint ( is_frequent, new_consequents );

    } // end of if

  } // END private void ap_genrules_constraint ( ARMinerItemset is_frequent, 
  //					Vector consequents ) 

//********************************************************//

  /**
   * Generates all the possible rules from a maximal
   * frequent itemset that meet all our criteria and adds
   * them to the rules.
   * 
   * @param itemset
   * a maximal frequent itemset from which to build rules
   */

  private void generateFromMaximal ( ARMinerItemset itemset )
  {
    // don't need this here any more, we do it in the calling method
    // -kap 2003-03-05
    // does the itemset have enough items to make rules?
    //    if ( itemset.size() < minConsequent + minAntecedent )
    //    {
    //      return;
    //    } // end if the itemset does not have enough items to

    // generate rules

    // simply go through all the sizes for the antecedent and consequent
    // that are possible. 
    int startAnt;
    int startCon;
    int stopAnt;
    int stopCon;

    // set start ant
    if ( minAntecedent == 0 )
    {
      startAnt = 1;
    } // end if no min specified
    else
    {
      startAnt = minAntecedent;
    } // end else min specified

    // set start con
    if ( minConsequent == 0 )
    {
      startCon = 1;
    } // end if no min specified
    else
    {
      startCon = minConsequent;
    } // end else min specified

    // set stop ant
    if ( maxAntecedent == 0 )
    {
      // leave enough items for the consequent
      stopAnt = itemset.size() - startCon;
    } // end if max not specified
    else
    {
      stopAnt = maxAntecedent;
    } // end else max specified

    // set stop con
    if ( maxConsequent == 0 )
    {
      // leave enough items for the antecdent
      stopCon = itemset.size() - startAnt;
    } // end if max not specified
    else
    {
      stopCon = maxConsequent;
    } // end else max specified

    int ant, con;

    switch ( debug ) {
    case 3: //case 0:
      System.out.println ( "AprioriRules:generateFromMaximal: " +
			   "itemset size= " + itemset.size() + " " +
			   "antecedent range: " + startAnt + "-" + stopAnt + 
			   " consequent range: " + startCon + "-" + stopCon );
    } // end switch debug 

    for ( ant = startAnt; ant <= stopAnt; ant++ )
    {
      for ( con = startCon; con <= stopCon; con++ )
      {
	// if we don't have enough items just skip it
	if ( con > itemset.size() - ant )
	{
	  break;
	} // end if not enough items left

	// generate rules with these sizes then
	generateFromMaximal ( itemset, ant, con );

      } // end for each consequent size

    } // end for each antecedent size

  } // END private void generateFromMaximal ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Generates all the possible rules from a maximal
   * frequent itemset that meet all our criteria and adds
   * them to the rules.
   * 
   * @param itemset
   * a maximal frequent itemset from which to build rules
   *
   * @param ant
   * the size to make the antecedent
   *
   * @param con
   * the size to make the consequent
   */

  private void generateFromMaximal ( ARMinerItemset itemset,
				     int ant, int con )
  {
    switch ( debug ) {
    case 1: //case 0:
      System.out.println ( "AprioriRules.generateFromMaximal: " +
			   "building rules with: " +
			   "itemset size= " + itemset.size() + " " +
			   "antecedent size= " + ant + " " +
			   "consequent size= " + con );
    } // end switch debug

    // need some indexes for ant and con
    int [] antIndexes = new int [ ant ];

    int i;

    // we should start our antecedent at the begining
    for ( i = 0; i < ant; i++ )
    {
      antIndexes [ i ] = i;
    } // for each ant index

    // ok, start with the first level
    generateFromMaximal ( itemset, con, antIndexes, 0 );

  } // END private void generateFromMaximal ( ARMinerItemset itemset,
  //				     int ant, int con )

//********************************************************//

  /**
   * Generates all the possible rules from a maximal
   * frequent itemset that meet all our criteria and adds
   * them to the rules.
   * 
   * @param itemset
   * a maximal frequent itemset from which to build rules
   *
   * @param con
   * the size to make the consequent
   *
   * @param antIndexes
   * indexes into itemset to use for the antecedent
   *
   * @param antLevel
   * the ant index to increment
   */

  private void generateFromMaximal ( ARMinerItemset itemset,
				     int con, int [] antIndexes, int antLevel )
  {
    switch ( debug ) {
    case 1: //case 0:
      System.out.print ( "AprioriRules.generateFromMaximal: " +
			 "antIndexes= " );
      for ( int i = 0; i < antIndexes.length; i++ )
      { System.out.print ( antIndexes [ i ] + " " ); }
      System.out.print ( "antLevel= " + antLevel + "\n" );
    } // end switch debug

    // ok, if we are incrementing the last ant level then 
    // we should generate the consequences, otherwise we should just 
    // call ourselves, increasing the ant level for each iteration
    // if this is the last level then we don't have to adjust the indexes
    // occuring after
    int generateConsequences = 0;

    if ( antLevel == antIndexes.length - 1 )
    {
      generateConsequences = 1;
    } // end if this is the last level

    int i;
    int [] antCopy = new int [ antIndexes.length ];
    int [] conIndexes = new int [ con ];

    for ( ; 
	  antIndexes [ antLevel ] <= 
	    itemset.size() - ( antIndexes.length - antLevel );
	  antIndexes [ antLevel ]++ )
    {
      // copy the indexes
      System.arraycopy ( antIndexes, 0, antCopy, 0, antIndexes.length );

      switch ( generateConsequences )
      {
      case 0: // do not generate consequences

	// generate the rest of the antecedents
	generateFromMaximal ( itemset, con, antCopy, antLevel + 1 );
	break;

      case 1: // generate consequences

	// set the con indexes to the first con indexes not used by
	// ant indexes
	int index = 0;
	for ( i = 0; i < con; i++ )
	{
	  while ( index < itemset.size() )
	  {
	    if ( Arrays.binarySearch ( antIndexes, index ) < 0 )
	    {
	      conIndexes [ i ] = index;
	      index++;
	      break;
	    } // end if this index is not used

	    // try the next index
	    index++;

	  } // end while there are still items to try

	} // end for each con index

	generateFromMaximal ( itemset, antIndexes, conIndexes, 0 );
	break;
      } // end switch make rule

      // have to increment each index after ant level
      for ( i = antLevel + 1; i < antIndexes.length; i++ )
      {
	antIndexes [ i ]++;
      } // end for each index after ant level

    } // end for each value of ant level index

  } // END private void generateFromMaximal ( ARMinerItemset itemset,
  //				     int con, 
  //                                 int [] antIndexes, int antLevel )

//********************************************************//

  /**
   * Generates all the possible rules from a maximal
   * frequent itemset that meet all our criteria and adds
   * them to the rules.
   * 
   * @param itemset
   * a maximal frequent itemset from which to build rules
   *
   * @param antIndexes
   * the antecedent indexes
   *
   * @param conIndexes
   * the consequent indexes
   *
   * @param conLevel
   * the consequent index to increment
   */

  private void generateFromMaximal ( ARMinerItemset itemset,
				     int [] antIndexes,
				     int [] conIndexes, int conLevel )
  {
    switch ( debug ) {
    case 1: //case 0:
      System.out.print ( "AprioriRules.generateFromMaximal: " +
			 "conIndexes= " );
      for ( int i = 0; i < conIndexes.length; i++ )
      { System.out.print ( conIndexes [ i ] + " " ); }
      System.out.print ( "conLevel= " + conLevel + "\n" );
    } // end switch debug

    // ok, if we are incrementing the last con level then 
    // we should make rules, otherwise we should just 
    // call ourselves, increasing the con level for each iteration
    // if this is the last level then we don't have to adjust the indexes
    // occuring after
    int makeRule = 0;

    if ( conLevel == conIndexes.length - 1 )
    {
      makeRule = 1;
    } // end if this is the last level

    int i;
    int [] conCopy = new int [ conIndexes.length ];
    int [] antCopy = new int [ antIndexes.length ];

    for ( ; 
	  conIndexes [ conLevel ] <= 
	    itemset.size() - ( conIndexes.length - conLevel );
	  conIndexes [ conLevel ]++ )
    {
      // if the current con index at con level is already in the
      // ant indexes skip it
      if ( Arrays.binarySearch ( antIndexes, conIndexes [ conLevel ] ) >= 0 )
      {
	continue;
      } // end if this index already used in ant indexes

      switch ( makeRule )
      {
      case 0: // do not make rules

	// copy the indexes
	System.arraycopy ( antIndexes, 0, antCopy, 0, antIndexes.length );
	System.arraycopy ( conIndexes, 0, conCopy, 0, conIndexes.length );
	// go generate the rest
	generateFromMaximal ( itemset, antCopy, conCopy, conLevel + 1 );
	break;

      case 1: // make rules

	buildRule ( itemset, antIndexes, conIndexes );

	switch ( debug ) {
	case 3: //case 0:
	  System.out.print ( "AprioriRules.generateFromMaximal: " +
			     "antIndexes=" ); 
	  for ( i = 0; i < antIndexes.length; i++ )
	  { System.out.print ( " " + antIndexes [ i ] ); }
	  System.out.print ( " conIndexes=" );
	  for ( i = 0; i < conIndexes.length; i++ )
	  { System.out.print ( " " + conIndexes [ i ] ); }
	  System.out.println ( "" );
	} // end switch debug

	break;
      } // end switch make rule

      // have to increment each index after con level
      for ( i = conLevel + 1; i < conIndexes.length; i++ )
      {
	conIndexes [ i ]++;

	// make sure that this index is not in use by ant indexes
	while ( conIndexes [ i ] < itemset.size() &&
		Arrays.binarySearch ( antIndexes, conIndexes [ i ] ) >= 0 )
	{
	  conIndexes [ i ]++;
	} // end while the index is in ant indexes

      } // end for each index after con level

    } // end for each value of con level index

  } // END private void generateFromMaximal ( ARMinerItemset itemset,
  //				     int [] antIndexes,
  //				     int [] conIndexes, int conLevel )

//********************************************************//

  /**
   * Builds a rule based on the specified itemset indexes
   * and adds it to the rules to be retuned. Note: confidence
   * is not calculated for rules whose antecedent and
   * consequent contains event items. The confidence of these
   * rules is the responsiblity of the calling method.
   *
   * @param itemset
   * a maximal frequent itemset from which to build rules
   *
   * @param antIndexes
   * the antecedent indexes
   *
   * @param conIndexes
   * the consequent indexes
   */

  private void buildRule ( ARMinerItemset itemset,
			   int [] antIndexes, 
			   int [] conIndexes )
  {
    // ok, here's a real subtle piece... since itemset is a maximally
    // frequent itemset and we may not be using all the items
    // to build the rule we must construct a temp itemset to retrieve
    // support for and give to the rule for reference instead of
    // using the itemset given us

    // for now we can simply add the antecedent and consequent
    // together

    // we could also merge the indexes, sort them and build
    // an itemset using buildItemset... hrm, this is probably a better
    // solution, we'll do that.
    int [] bothIndexes = new int [ antIndexes.length + conIndexes.length ];

    System.arraycopy ( antIndexes, 0, bothIndexes, 0, antIndexes.length );
    System.arraycopy ( conIndexes, 0, bothIndexes, antIndexes.length, 
		       conIndexes.length );

    Arrays.sort ( bothIndexes );

    ARMinerItemset tempItemset = 
      ARMinerItemset.buildItemsetFromIndexes ( itemset, bothIndexes );

    ARMinerItemset antecedent = 
      ARMinerItemset.buildItemsetFromIndexes ( itemset, antIndexes );

    ARMinerItemset consequent = 
      ARMinerItemset.buildItemsetFromIndexes ( itemset, conIndexes );

    // make sure the confidence is high enough

    // Here's another tricky part. For rules that have events in the
    // antecedent and the consequent of the rule we have to consider
    // that events may show up many times in a single instance. What 
    // happens when we have a rule A => B, where A and B are events.
    // Let's assume A happens before B. A single instance could have
    // many A's. Conventionally, if any A is followed by a B we count
    // that instance towards the support of the itemset AB. If every
    // instance containing an A has at least one A followed by a B
    // the confidence of A => B is 100%. The problem is that every A
    // in a single instance might not be followed by a B. Hence, our
    // data does not represent a 100% confidence. What we have to do
    // is consider each A in a single instance and each AB. Using the
    // number of times these appear in our data (the Event Weight), 
    // which could be more than the number of instances, to calculate
    // confidence should result in a much more accurate view.

    // to save time the counting of occurrances of events are delayed
    // until after rule generation. for rules whose antecedent and
    // consequent contain event we will just set the confidence
    // to 1. The calling method will be responsible for correctly 
    // calculating that rules confidence. -kap 2003-12-11

    tempItemset.setSupport ( (float)0.000001 );
    antecedent.setSupport ( (float)0.000001 );
    consequent.setSupport ( (float)0.000001 );

    // yeah yeah, I'm checking it
    float confidence =  0;

    try 
    {
      // get suports
      tempItemset.setSupport ( supports.getSupport ( tempItemset ) );

      // put this off for the calling method to do -kap 2003-12-11
      
      // get the event weights too
      /*
	tempItemset.setEventWeight ( supports.getEventWeight ( tempItemset ) );
	antecedent.setEventWeight ( supports.getEventWeight ( antecedent ) );
	consequent.setEventWeight ( supports.getEventWeight ( consequent ) );
      */

    } // end try getting supports
    catch ( SETException e ) 
    {
      System.err.println ( "AprioriRules.buildRule: " +
			   "Error getting rule support from SET.\n\t" + e );
    } // end catch set exception

    try 
    {
      // get suports
      antecedent.setSupport ( supports.getSupport ( antecedent ) );
    } // end try getting supports
    catch ( SETException e ) 
    {
      System.err.println ( "AprioriRules.buildRule: " +
			   "Error getting antecedent support from SET.\n\t" + 
			   e );
    } // end catch set exception

    try 
    {
      // get suports
      consequent.setSupport ( supports.getSupport ( consequent ) );
    } // end try getting supports
    catch ( SETException e ) 
    {
      System.err.println ( "AprioriRules.buildRule: " +
			   "Error getting consequent support from SET.\n\t" + 
			   e );
    } // end catch set exception

    // if both antecedent and consequent have at least one event each
    // use the event weight instead of support

    // instead we will just set confidence to 1. -kap 2003-12-11

    if ( antecedent.numEventItems() > 0 && consequent.numEventItems() > 0 )
    {
      confidence = (float)0.95;

      // calculate confidence, cast as floats so the math is done nice
      /*
	confidence = ((float)tempItemset.getEventWeight()) / 
	((float)antecedent.getEventWeight());

      switch ( debug ) {
      case 3: //case 0:
	System.out.println ( "AprioriRules.buildRule: " +
			     "Used event weight. " +
			     "antEW = " + antecedent.getEventWeight() +
			     " itemsetEW = " + tempItemset.getEventWeight() +
			     " confidence= " +
			     confidence +
			     " would have been= " +
			     tempItemset.getSupport() / 
			     antecedent.getSupport() ); 
      } // end switch debug
      */

    } // end if ant and con have events
    else
    {
      // calculate confidence
      confidence = tempItemset.getSupport() / antecedent.getSupport();

    } // end else use regular support
    
    if ( confidence < minConfidence )
    {
      // don't do it!!!
      return; 
    } // end if confidence sucks

    // make sure each has the required items
    if ( requiredAntecedents != null &&
	 ! antecedent.hasAttributes ( requiredAntecedents ) )
    {
      return;
    } // end if required antecedents not there

    if ( requiredConsequents != null &&
	 ! consequent.hasAttributes ( requiredConsequents ) )
    {
      return;
    } // end if required consequents not there

    // add the rule fool
    rules.add ( new AssociationRule ( antecedent, 
				      consequent,
				      tempItemset,  // the original
				      tempItemset.getSupport(),
				      confidence ) );    

    // set the number of instances so neat statistical things
    // like chi-square can be calculated
    ((AssociationRule)rules.lastElement()).
      setNumInstances ( numInstances );
    
  } // END private void buildRule ( ARMinerItemset itemset,
  //			   int [] antIndexes, 
  //			   int [] conIndexes )

//********************************************************//

  /**
   * Removes duplicate rules from the rules list. The number
   * of rules removed is returned.
   *
   * @return the number of duplicate rules found and removed
   */

  private int removeDuplicateRules()
  {
    // count the number of duplicate rules found and removed
    int numRemoved = 0;

    // compare each rule
    for ( int i = 0; i < rules.size(); i++ )
    {
      numRemoved += removeDuplicateRules ( i );
      
    } // end for each rule

    // all done, return how many we removed
    return ( numRemoved );

  } // END private int removeDuiplicateRules()

//********************************************************//

  /**
   * Removes duplicate rules from the rules list. The rule
   * at the specified index is compared to all rules
   * appearing later in the list of rules. The number
   * of rules removed is returned.
   *
   * @param index
   * the index of the rule list where comparisons should 
   * start
   *
   * @return the number of duplicate rules found and removed
   */

  private int removeDuplicateRules ( int index )
  {
    // count the number of duplicate rules found and removed
    int numRemoved = 0;

    // compare the rule at index to the rules appearing
    // later in the list
    for ( int i = index + 1; i < rules.size(); i++ )
    {
      if ( ((AssociationRule)rules.elementAt ( index )).equals 
	   ( (AssociationRule)rules.elementAt ( i ) ) )
      {
	// remove the duplicate rule
	rules.removeElementAt ( i );
	
	// decrement i so we don't accidentally skip any
	// rules
	i--;

	// increment the num removed counter
	numRemoved++;

      } // end if the rules are equal
      
    } // end for each rule

    // all done, return how many we removed
    return ( numRemoved );

  } // END private int removeDuiplicateRules()

//********************************************************//

} // END public class AprioriRules implements AssociationsFinder 
