/*
ARMiner - Association Rules Miner
Copyright (C) 2000  UMass/Boston - Computer Science Department

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The ARMiner Server was written by Dana Cristofor and Laurentiu
Cristofor.

The ARMiner Client was written by Abdelmajid Karatihy, Xiaoyong Kuang,
and Lung-Tsung Li.

The ARMiner package is currently maintained by Laurentiu Cristofor
(laur@cs.umb.edu).
*/

package wpi.associations.arminerSequence;

import java.util.*;

// stuff for dealing with time sequence stuff
import wpi.timesequence.*;

/*
   This file is a part of the ARMiner project.
   
   (P)1999-2000 by ARMiner Server Team:

   Laurentiu Cristofor
   Lung-Tsung Li

   modified by Keith A. Pray (kap@wpi.edu)
*/
	
/*
  Maintainance log started on November 30th, 2000

  Nov. 30th, 2000   - added getItemsets method
                    - added and renamed some private methods
                    - improved toString method

  Apr. 30th, 2002   - added support for time sequence attributes
                    - added debug info
		    - cleaned up formatting -kap

  May  27th, 2002   - clarified exception messages -kap

  June 1, 2002      - added comments to code so it is easy to follow.
                      please forgive me if I err. 
		      Also added indexing for time sequence events -kap

  January 20, 2003  - added event weight to supplement support
                      for itemsets containing time sequence events -kap

  February 24, 2003  - made consistent the convention of using a lower case
		       letter to start a method name. -kap
*/

/**
 * Implements a Set Enumeration Tree, which is a prefix tree used
 * for storing and retrieving itemset information.
 *
 * @author Dana Cristofor
 * @author Laurentiu Cristofor
 * @author Keith A. Pray (kap@wpi.edu)
 */

public class SET
{

//********************************************************//
// ----- Inner Classes -----
//********************************************************//  

  private class HashNode
  {
    //********************************************************//

    Hashtable children;
    float support;

    // needed for time sequence support -kap

    /** The event weight. The number of times the events in an
     *  itemset appear in the dataset this item was found
     *  frequent in. 
     */
    long eventWeight;

    /** The name of the item represented by this node if it's
     *  a time sequence event. Null otherwise
     */
    String event;

    //********************************************************//    

    public HashNode()
    {
      support = 0;
      eventWeight = 0;
      children = new Hashtable();

      // don't know the name yet, if any -kap
      event = null;
    }

    //********************************************************//    

    public String toString()
    {
      // changed to use string buffer -kap
      StringBuffer sb = new StringBuffer ( "<children: " );

      sb.append ( children.toString() );
      sb.append ( " support: " );
      sb.append ( support );
      sb.append ( " event weight: " );
      sb.append ( eventWeight );
      //      sb.append ( ">\n" );
      sb.append ( ">" );

      return ( sb.toString() );
    }

    //********************************************************//    

  } // END private class HashNode

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  private HashNode root;
  private int level = 0;

  // ----- Stuff needed for time sequence attribute support. -kap ----- //
  
  /** Used to look up actual value of an item using its 
      integer representation. */
  private Hashtable numberHash;

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: and out from methods
   * 3: and all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Sets the number hash to use for looking up values
   * in this itemset
   *
   * @param h
   * the number hash
   */

  public void setNumberHash ( Hashtable h )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.println ( "SET.setNumberHash: size = " + h.size() );
    } // end switch debug -kap

    numberHash = h;

  } // END public void setNumberHash ( Hashtable h )

//********************************************************//

  /**
   * Returns the number hash used for looking up values
   * in this itemset.
   *
   * @return the number hash
   */

  public Hashtable getNumberHash()
  {
    return ( numberHash );

  } // END public Hashtable getNumberHash()

//********************************************************//
  
  /**
   * Create a new empty SET.
   */

  public SET()
  {
    root = new HashNode();

  } // END public SET()

//********************************************************//
  
  /**
   * Insert a new itemset in the SET.
   *
   * @param itemset  
   * the itemset to be inserted
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null or is empty
   */

  public void insert ( ARMinerItemset itemset )
  {
    if ( itemset == null || itemset.size() == 0 )
    {
      throw new IllegalArgumentException 
	( "SET.insert: itemset must be non null and non empty" );
    }

    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "SET.insert: Making new itemset." );
    case 1: //case 0:
      System.out.println ( "SET.insert: itemset = " + itemset );
    } // end switch debug -kap

    // doesn't seem to be much need for creating a new itemset here
    // maybe they were worried about upsetting the internal
    // position of the enumeration type functionality, but I don't
    // see it used many places. -kap
    ARMinerItemset is = new ARMinerItemset ( itemset );

    // it is not neccessary to set the number hash here since the 
    // constructor should get it from itemset -kap

    // but let's compare this itemset's number hash to the one we have
    // if they are the same, no worries
    // do we have a number hash?
    if ( numberHash == null )
    {
      // then get a ref to this number hash.. better than none
      numberHash = is.getNumberHash();
    } // end if we don't have a number hash
    else
    {
      // are they the same? if not we better compare
      if ( ! numberHash.equals ( is.getNumberHash() ) )
      {
	// who has more?
	if ( numberHash.size() < is.getNumberHash().size() )
	{
	  // well, I'm gonna guess that itemset's is better to have
	  numberHash = is.getNumberHash();

	} // end if our number hash is bigger

      } // end if number hashes aren't the same

    } // end else we have one but should compare it to the itemset's

    Object obj;
    HashNode node;
    HashNode walker = root;
    Integer key;

    // if an item is an event there will be another node we have to deal with
    HashNode eventNode = null;
    // and the key will be a string
    String eventKey;
    // to build the event key quickly
    StringBuffer sb = new StringBuffer();

    // for each item in the itemset walk the branch of the tree
    // associated with it -kap
    while ( is.hasMoreItems() )
    {
      // the key for looking up the next node in the tree is the next
      // item in the itemset -kap
      key = new Integer ( is.getNextItem() );
      
      // check if this item is an event
      if ( is.eventName ( key ) == null )
      {
	eventKey = null;
      }
      else
      {
	// better build the key for it then
	sb.setLength ( 0 );
	sb.append ( is.relativeBegin ( key ) );
	sb.append ( TSUtility.TIME_SEQUENCE_DELIMITER );
	sb.append ( is.relativeEnd ( key ) );
	eventKey = sb.toString();

      } // end else item is an event

      // if there is already a node for this item go to it -kap
      if ( ( obj = walker.children.get ( key ) ) != null )
      {
	walker = (HashNode)obj;

	// if it was an event we have to check if the event key node exists
	if ( eventKey != null )
	{
	  // does the event node exist?
	  if ( ( eventNode = (HashNode)walker.children.get ( eventKey ) ) 
	       != null )
	  {
	    // then we walk to the event node
	    walker = eventNode;
	    
	    switch ( debug ) {
	    case 3:
	      System.out.println ( "SET.insert: walked to event node " + key +
				   " : " + eventKey + " : " + walker );
	    } // end switch debug

	  } // end if event node exists
	  else
	  {
	    // we had better build a node for the event then
	    eventNode = new HashNode();

	    // add this new event node with the event key
	    walker.children.put ( eventKey, eventNode );
	    
	    // now walk to it baby
	    walker = eventNode;
	    
	  } // end else we have to make the event node

	} // end if the current item is an event
	else
	{
	  // then we can walk right to the regular node
	  walker = (HashNode)obj;
	  
	  switch ( debug ) {
	  case 3:
	    System.out.println ( "SET.insert: walked to regular node " + key +
				 " : " + walker );
	  } // end switch debug
	  
	} // end else it was a regular item
	
      } // end if we found the node for the item
      else
      {
	// otherwise we must make a new node for this item -kap
	node = new HashNode();
	walker.children.put ( key, node );
	walker = node;
	
	// is the item an event? 
	if ( eventKey != null )
	{
	  // we also have to make an event node then

	  // first assign the event name to the node
	  node.event = is.eventName ( key );

	  // well, now we have to make another node and add it to
	  // the children of the node we just made above
	  eventNode = new HashNode();

	  // add this node with the event key
	  node.children.put ( eventKey, eventNode );

	  // now walk to it baby
	  walker = eventNode;
	  
	  switch ( debug ) {
	  case 3:
	    System.out.println ( "SET.insert: added an event node for: " +
				 node.event + " with key: " + key + " - " +
				 eventKey + " in node: " + node );
	  } // end switch debug
	  
	} // end if item is an event
	
	// since we know there are no other nodes from here on
	// we could save time bye just creating nodes for the rest
	// of the items rather than checking above. But will we?
	// not yet I guess. -kap

      } // end else we have to make a new node

    } // end while we have more items left

    // assign the current node the support of the itemset we just
    // used to walk the tree -kap
    walker.support = is.getSupport();

    // remember to set the event weight of the itemset
    // this is used to track the number of times the event items
    // appear in a dataset. this can be more than the number
    // of instances in the dataset originally mined.
    walker.eventWeight = is.getEventWeight();
    
    // this is a big problem for supporting time sequence stuff.
    // since the itemset is not actually stored in the tree and
    // the tree must be walked using the item numbers to find 
    // the support I don't see any easy way to store the support
    // for itemsets that have the same items but different
    // relative times. -kap
    
    // Ooo, since I never change the relative begin and end time value
    // of an item once it is added to an itemset I may be able to
    // use the begin:end as an additional key for items that are
    // events. As in, a node would be made for each item just as
    // things currently work, but when an item is an event all the
    // nodes directly adjacent will have begin:end for keys. -kap
    
    switch ( debug ) {
    case 3: case 2: //case 0:
      System.out.println ( "SET.insert: " +
			   "assigned support and event weight to node = " +
			   walker );
    } // end switch debug -kap
    
  } // END public void insert ( ARMinerItemset itemset )

//********************************************************//
  
  /**
   * Return the support for a given itemset.
   *
   * @param itemset  
   * the itemset for which we want to obtain the support
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null or is empty
   *
   * @exception SETException  
   * <code>itemset</code> not found in SET
   *
   * @return support
   */

  public float getSupport ( ARMinerItemset itemset )
    throws SETException
  {
    // get the node for the itemset specified
    HashNode node = getNode ( itemset );

    // now we just return the support for that node -kap
    return ( node.support );

  } // END public float getSupport ( ARMinerItemset itemset )

//********************************************************//
  
  /**
   * Return the event weight for a given itemset.
   *
   * @param itemset  
   * the itemset for which we want to obtain the event weight
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null or is empty
   *
   * @exception SETException  
   * <code>itemset</code> not found in SET
   *
   * @return event weight
   */

  public long getEventWeight ( ARMinerItemset itemset )
    throws SETException
  {
    // get the node for the itemset specified
    HashNode node = getNode ( itemset );

    // now we just return the event weight for that node -kap
    return ( node.eventWeight );

  } // END public long getEventWeight ( ARMinerItemset itemset )

//********************************************************//
 
  /**
   * Return the node for a given itemset.
   *
   * @param itemset  
   * the itemset for which we want to obtain the node
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null or is empty
   *
   * @exception SETException  
   * <code>itemset</code> not found in SET
   *
   * @return the node representing the itemset
   */

  private HashNode getNode ( ARMinerItemset itemset )
    throws SETException
  {
    if ( itemset == null || itemset.size() == 0 )
    {
      throw new IllegalArgumentException
	( "SET.getNode(): aurgument must be non null and non empty" );
    } // end if itemset is stupid -kap

    Object obj;

    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "SET.getNode: Making new itemset" );
    case 1:
      System.out.print ( "SET.getNode: of itemset: " + itemset );
    } // end switch debug -kap

    // why is a new itemset made here? It just seems wastefull -kap

    ARMinerItemset is = new ARMinerItemset ( itemset );

    // it is not neccessary to set the number hash here since the 
    // constructor should get it from itemset -kap

    // start walking the tree at the root of course -kap
    HashNode walker = root;
    Integer key;

    // might need a string buffer to build event keys quickly
    StringBuffer sb = new StringBuffer();

    HashNode eventNode;

    // walk to the appropriate node for each item in the itemset -kap
    while ( is.hasMoreItems() )
    {
      // the key is the item iteger representation -kap
      key = new Integer ( is.getNextItem() );
      
      // if there is no node for this item then this itemset will not
      // be found in the tree -kap
      if ( ( obj = walker.children.get ( key ) ) == null )
      {
	throw new SETException ( "SET.getNode: itemset " +
				 itemset + " not found in SET! " +
				 "Stopped on item " + key );
      } // end if we're never gonna find the itemset

      // if this item is an event we need to use the begin:end time
      // key in addition to the item number
      if ( is.eventName ( key ) != null )
      {
	// Build the event key for it
	sb.setLength ( 0 );
	sb.append ( is.relativeBegin ( key ) ).
	  append ( TSUtility.TIME_SEQUENCE_DELIMITER ).
	  append ( is.relativeEnd ( key ) );

	eventNode = (HashNode)obj;

	switch ( debug ) {
	case 3: //case 0:
	  System.out.println ( "SET.getNode: walking to eventNode " +
			       key + " : " +
			       eventNode );
	} // end switch debug

	// make sure it's there
	if ( ( obj = eventNode.children.get ( sb.toString() ) ) == null )
	{
	  throw new SETException ( "SET.getNode: itemset " +
				   itemset + " not found in SET! " +
				   "Stopped on item " + key + " : " +
				   sb.toString() );
	} // end if we're never gonna find the itemset
	
      } // end if the item is an event
      
      // move to the event node -kap
      walker = (HashNode)obj;

    } // end while more items -kap

    switch ( debug ) {
    case 3: case 2: //case 0:
      System.out.println ( "SET.getNode: Found node = " + 
			   walker );
    } // end switch debug -kap

    // now we just return the node
    return ( walker );

  } // END private HashNode getNode ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Return the maximal itemsets of the SET. 
   *
   * @return
   * a vector containing the maximal itemsets from the SET
   */

  public Vector getMaximalItemsets() 
  { 
    Vector v = new Vector();

    if ( ! root.children.isEmpty() )
    {
      switch ( debug ) {
      case 3: case 2:
	System.out.println ( "SET.getLargeItemsets: Making new itemset." );
      } // end switch debug -kap

      // we must pass the number hash
      // to the new itemset to support time sequence
      // attributes -kap
      ARMinerItemset newItemset = new ARMinerItemset();
      newItemset.setNumberHash ( getNumberHash() );

      traverseGatherLeaves ( root, newItemset, v );

    } // end if the tree isn't empty

    ARMinerItemset.pruneNonMaximal ( v );

    return ( v );

  } // END public Vector getMaximalItemsets() 

//********************************************************//

  /**
   * Return the itemsets of the SET. 
   *
   * @return
   * a vector containing the itemsets from the SET
   */

  public Vector getItemsets() 
  { 
    Vector v = new Vector();

    if ( ! root.children.isEmpty() )
    {
      switch ( debug ) {
      case 3: case 2:
	System.out.println ( "SET.getItemsets: Making new itemset." );
      } // end switch debug -kap

      // we must pass the number hash
      // to the new itemset to support time sequence
      // attributes -kap
      ARMinerItemset newItemset = new ARMinerItemset();
      newItemset.setNumberHash ( numberHash );

      traverseGatherAll ( root, newItemset, v );

    } // end if root node children not empty

    return ( v );
  } // END public Vector getItemsets() 

//********************************************************//

  /**
   * A private method which gets called recursively to retrieve itemsets 
   * from the leaf nodes of the SET.
   *
   * @param node
   * node starts from the root node of the SET. 
   *
   * @param itemset
   * for storing hashtable keys as it traverses node by 
   * node to the leaf. 
   *
   * @param vector
   * for storing itemsets . 
   */

  private void traverseGatherLeaves ( HashNode node, 
				      ARMinerItemset itemset, 
				      Vector vector )
  {
    // if the children of this node is empty it is a leaf -kap
    if ( node.children.isEmpty() )
    {
      switch ( debug ) {
      case 3: case 2:
	System.out.println ( "SET.traverseGatherLeaves: Making new itemset." );
      } // end switch -kap

      // need to make a new itemset now since we are growing all the
      // itemsets represent by this SET tree
      ARMinerItemset is = new ARMinerItemset ( itemset );
      
      // it is not neccessary to set the number hash here since the 
      // constructor should get it from itemset -kap
      
      is.setSupport ( node.support );
      is.setEventWeight ( node.eventWeight );
      vector.addElement ( is );
      return;

    } // end if node children empty

    Enumeration e = node.children.keys();
    Integer key;
    HashNode childNode;

    while ( e.hasMoreElements() )
    {
      key = (Integer)e.nextElement();
      childNode = (HashNode)node.children.get ( key );

      // is it a regular item node?
      if ( childNode.event == null )
      {
	// add the regular item info
	itemset.addRegularItem ( key.intValue() );
	
	// it is not neccessary to set the number hash here since the 
	// constructor should get it from itemset -kap
	
	traverseGatherLeaves ( childNode, itemset, vector );

	// and pretend we didn't add that last item so it doesn't
	// show up in other walks of the tree -kap
	itemset.removeLastItem();
	
      } // end if node is for regular item
      else
      {
	// then we have to do all sorts of tricky stuff
	Enumeration eventE = childNode.children.keys();

	// we should be getting string keys
	String eventKey;

	// the begin and end times of the event
	double [] times;

	while ( eventE.hasMoreElements() )
	{
	  // first add the regular info
	  itemset.addRegularItem ( key.intValue() );

	  // get the key for the next node
	  eventKey = (String)eventE.nextElement();

	  // get the values of the begin and end times for the node
	  times = TSUtility.getValues ( eventKey );

	  // add the event info to itemset, the item number, the name,
	  // the begin and end time
	  itemset.addEventItem ( key.intValue(), childNode.event,
				 times [ 0 ], times [ 1 ] );

	  // now we continue traversing the tree with the
	  // event key node
	  traverseGatherLeaves ( (HashNode)childNode.children.get ( eventKey ),
				 itemset, vector );

	  // and pretend we didn't add that last item so it doesn't
	  // show up in other walks of the tree -kap
	  itemset.removeLastItem();

	} // end while event E has more

      } // end else the node is for an event item

    } // end while e has more

  } // END private void traverseGatherLeaves ( HashNode node, 
  //				      ARMinerItemset itemset, 
  //				      Vector vector )

//********************************************************//
  
  /**
   * A private method which gets called recursively to retrieve itemsets 
   * from all the nodes of the SET.
   *
   * @param node
   * node starts from the root node of the SET. 
   *
   * @param itemset
   * for storing hashtable keys as it traverses node by 
   * node to the leaf. 
   *
   * @param vector
   * for storing itemsets . 
   */

  private void traverseGatherAll ( HashNode node, ARMinerItemset itemset, 
				   Vector vector )
  {
    if ( node.children.isEmpty() )
    {
      return;
    }

    Enumeration e = node.children.keys();
    Integer key;
    HashNode childNode, tempNode;
    ARMinerItemset is;

    while ( e.hasMoreElements() )
    {
      key = (Integer)e.nextElement();
      childNode = (HashNode)node.children.get ( key );

      // is it a regular item node?
      if ( childNode.event == null )
      {
	// add the regular item info
	itemset.addRegularItem ( key.intValue() );

	// each node visited represents another itemset stored in
	// this tree
	is = new ARMinerItemset ( itemset );
	
	// it is not neccessary to set the number hash here since the 
	// constructor should get it from itemset -kap
	
	is.setSupport ( childNode.support );
	is.setEventWeight ( childNode.eventWeight );
	vector.addElement ( is );
	
	traverseGatherAll ( childNode, itemset, vector );

	// and pretend we didn't add that last item so it doesn't
	// show up in other walks of the tree -kap
	itemset.removeLastItem();
	
      } // end if node is for regular item
      else
      {
	// then we have to do all sorts of tricky stuff
	Enumeration eventE = childNode.children.keys();

	// we should be getting string keys
	String eventKey;

	// the begin and end times of the event
	double [] times;

	while ( eventE.hasMoreElements() )
	{
	  // first add the regular info
	  itemset.addRegularItem ( key.intValue() );

	  // get the key for the next node
	  eventKey = (String)eventE.nextElement();

	  // get the values of the begin and end times for the node
	  times = TSUtility.getValues ( eventKey );

	  // add the event info to itemset, the item number, the name,
	  // the begin and end time
	  itemset.addEventItem ( key.intValue(), childNode.event,
				 times [ 0 ], times [ 1 ] );

	  // each node visited represents another itemset stored in
	  // this tree
	  is = new ARMinerItemset ( itemset );

	  // remember to set the support, have to get the support from 
	  // the node pointed to by eventKey... need event weight too
	  tempNode = (HashNode)childNode.children.get ( eventKey );

	  is.setSupport ( tempNode.support );
	  is.setEventWeight ( tempNode.eventWeight );

	  // add the itemset to
	  vector.addElement ( is );

	  // now we continue traversing the tree with the temp node
	  // event key node
	  traverseGatherAll ( tempNode, itemset, vector );

	  // and pretend we didn't add that last item so it doesn't
	  // show up in other walks of the tree -kap
	  itemset.removeLastItem();

	} // end while event E has more

      } // end else the node is for an event item

    } // end while e has more

  } // END private void traverseGatherAll ( HashNode node, 
  //				   ARMinerItemset itemset, 
  //				   Vector vector )

//********************************************************//
  
  /**
   * A private method which gets called recursively to
   * retrieve itemsets from each node of the SET and print
   * them to a String.
   *
   * @param node  
   * node to traverse
   *
   * @param s  
   * StringBuffer in which we store the representation
   * of the nodes we saw so far
   */

  private void traversePrint ( HashNode node, StringBuffer sb )
  {
    level++;

    if ( node.children.isEmpty() )
    {
      level--;
      return;
      
    } // end if node children empty

    switch ( debug ) {
    case 3:
      System.out.println ( "SET.traversePrint: children = " + node.children );
    } // end switch debug

    Enumeration e = node.children.keys();

    // key for regular items
    Integer key;

    // key for event items
    String eventKey;

    // temp node for checking for event node
    HashNode tempNode, tempNode2;

    while ( e.hasMoreElements() )
    {
      key = (Integer)e.nextElement();

	// make the nice indenting -kap
	for ( int i = 1; i < level; i++ )
	{
	  sb.append ( "   " );
	}
	
	// check if the node for key is for an event item
	tempNode = (HashNode)node.children.get ( key );
      
	// changed to use all appends instead of + string operator -kap
	sb.append ( "<" );
	sb.append ( key );
	sb.append ( ">:[" );
	sb.append ( level );
	sb.append ( "] support= " );
	sb.append ( tempNode.support );
	sb.append ( " event weight= " );
	sb.append ( tempNode.eventWeight );
	sb.append ( "\n" );

      // is this temp node for an event item?
      if ( tempNode.event == null )
      {
	traversePrint ( tempNode, sb );
      } // end if temp node for a regular item
      else
      {
	// we have to print the event info for each child node in temp
	// node
	Enumeration eventE = tempNode.children.keys();

	while ( eventE.hasMoreElements() )
	{
	  // make the nice indenting -kap
	  for ( int i = 1; i < level; i++ )
	  {
	    sb.append ( "   " );
	  }

	  // get the key for the next node
	  eventKey = (String)eventE.nextElement();

	  // get the node
	  tempNode2 = (HashNode)tempNode.children.get ( eventKey );

	  sb.append ( "<" );
	  sb.append ( eventKey );
	  sb.append ( ">:[" );
	  sb.append ( level );
	  sb.append ( "] support= " );
	  sb.append ( tempNode2.support );
	  sb.append ( " event weight= " );
	  sb.append ( tempNode2.eventWeight );
	  sb.append ( "\n" );

	  // now we continue traversing the tree with the temp node
	  // event key node
	  traversePrint ( tempNode2, sb );

	} // end while event E has more
	
      } // end else temp node is for an event item

    } // end while e has more
    
    level--;
    
  } // END private void traversePrint ( HashNode node, StringBuffer sb )

//********************************************************//
	
  /**
   * Return a string representation of the SET.
   *
   * @return string representation of SET
   */

  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    traversePrint ( root, sb );
    return ( sb.toString() );

  } // END public String toString()

//********************************************************//
  
  /** 
   * Main method for testing purposes.
   */

  public static void main ( String[] args )
  {
    // to test events and stuff
    Hashtable nh = new Hashtable();

    SET set = new SET();
    set.setNumberHash ( nh );

    ARMinerItemset is1 = new ARMinerItemset();
    is1.setNumberHash ( nh );
    ARMinerItemset is2 = new ARMinerItemset ( is1 );
    ARMinerItemset is3 = new ARMinerItemset ( is1 );
    ARMinerItemset is4 = new ARMinerItemset ( is1 );
    ARMinerItemset is5 = new ARMinerItemset ( is1 );
    ARMinerItemset is6 = new ARMinerItemset ( is1 );
    ARMinerItemset is7 = new ARMinerItemset ( is1 );
    ARMinerItemset is8 = new ARMinerItemset ( is1 );
    ARMinerItemset is9 = new ARMinerItemset ( is1 );

    nh.put ( new Integer ( 1 ), "odd-Event=1:10" );
    //nh.put ( new Integer ( 1 ), "odd-1" );
    is1.addItem ( 1 );
    //nh.put ( new Integer ( 2 ), "even-Event=2:20" );
    nh.put ( new Integer ( 2 ), "even-2" );
    is1.addItem ( 2 );
    nh.put ( new Integer ( 3 ), "odd-Event=3:30" );
    //nh.put ( new Integer ( 3 ), "odd-3" );
    is1.addItem ( 3 );
    //nh.put ( new Integer ( 4 ), "even-Event=4:40" );
    nh.put ( new Integer ( 4 ), "even-4" );
    is1.addItem ( 4 );
    is1.setSupport ( (float)0.4 );

    is2.addItem ( 1 );
    is2.addItem ( 2 );
    is2.addItem ( 3 );
    nh.put ( new Integer ( 7 ), "odd-Event=7:70" );
    //nh.put ( new Integer ( 7 ), "odd-7" );
    is2.addItem ( 7 );
    is2.setSupport ( (float)0.3 );

    is3.addItem ( 1 );
    is3.addItem ( 2 );
    is3.addItem ( 3 );
    nh.put ( new Integer ( 5 ), "odd-Event=5:50" );
    //nh.put ( new Integer ( 5 ), "odd-5" );
    is3.addItem ( 5 );
    //    nh.put ( new Integer ( 6 ), "even-Event=6:60" );
    nh.put ( new Integer ( 6 ), "even-6" );
    is3.addItem ( 6 );
    is3.setSupport ( (float)0.5 );

    is4.addItem ( 1 );
    is4.addItem ( 2 );
    is4.addItem ( 4 );
    is4.setSupport ( (float)0.65 );

    is5.addItem ( 1 );
    is5.addItem ( 2 );
    is5.addItem ( 5 );
    is5.setSupport ( (float)0.6 );

    is6.addItem ( 2 );
    is6.addItem ( 4 );
    is6.addItem ( 5 );
    is6.setSupport ( (float)0.55 );

    is7.addItem ( 1 );
    is7.addItem ( 2 );
    is7.setSupport ( (float)0.2 );

    is8.addItem ( 2 );
    is8.addItem ( 4 );
    is8.addItem ( 7 );
    //    nh.put ( new Integer ( 8 ), "even-Event=8:80" );
    nh.put ( new Integer ( 8 ), "even-8" );
    is8.addItem ( 8 );
    is8.setSupport ( (float)0.52 );

    is9.addItem ( 2 );
    is9.addItem ( 4 );
    is9.addItem ( 6 );
    is9.setSupport ( (float)0.59 );
    
    set.insert ( is2 );
    set.insert ( is3 );
    set.insert ( is1 );
    set.insert ( is4 );
    set.insert ( is5 );		
    set.insert ( is6 );
    set.insert ( is7 );
    set.insert ( is8 );
    set.insert ( is9 );
    /*
    Hashtable map = is1.canCombineWith ( is2 );

    //    System.out.println ( "\nis1 can combine with is2= " + map );

    Vector combined = is1.combineWith ( is2, map );

    //    System.out.println ( "combined = " + combined );

    set.insert ( (ARMinerItemset)combined.get ( 0 ) );
    */
    System.out.println ( set );
    
    try
    {
      System.out.println ( "Support for is1: " + set.getSupport ( is1 ) );
      System.out.println ( "Support for is2: " + set.getSupport ( is2 ) );
      System.out.println ( "Support for is3: " + set.getSupport ( is3 ) );
      System.out.println ( "Support for is4: " + set.getSupport ( is4 ) );
      System.out.println ( "Support for is5: " + set.getSupport ( is5 ) );
      System.out.println ( "Support for is6: " + set.getSupport ( is6 ) );
      System.out.println ( "Support for is7: " + set.getSupport ( is7 ) );
      System.out.println ( "Support for is8: " + set.getSupport ( is8 ) );
      System.out.println ( "Support for is9: " + set.getSupport ( is9 ) );
    }
    catch ( SETException e )
    {
      System.out.println ( e );
    }
    /*
    Vector v = set.getLargeItemsets();
    System.out.println ( "\nLarge itemsets are: " + '\n' );
    */

    System.out.println ( "We should get itemsets:" );
    System.out.println ( "is1: " + is1 );
    System.out.println ( "is2: " + is2 );
    System.out.println ( "is3: " + is3 );
    System.out.println ( "is4: " + is4 );
    System.out.println ( "is5: " + is5 );
    System.out.println ( "is6: " + is6 );
    System.out.println ( "is7: " + is7 );
    System.out.println ( "is8: " + is8 );
    System.out.println ( "is9: " + is9 );

    Vector v = set.getItemsets();
    System.out.println ( "\nThe itemsets are: " + '\n' );

    for ( int i = 0; i < v.size(); i++ )
    {
      System.out.println ( v.get ( i ) );
    }

    System.out.println ( "Maximal itemsets: " + set.getMaximalItemsets() );

  } // END public static void main ( String[] args )

//********************************************************//

} // END public class SET
