package wpi.filters;

import java.io.*;
import java.util.*;

import wpi.timesequence.*;

import weka.core.*;
import weka.filters.*;

/*
  Maintenance log started on March 20, 2004

  March 20, 2004     - created -kap

  April 22, 2004     - fixed listOptions to be accurate, still had info
                       from class this class was copied from.
		     - spell checked everything. -kap
*/

/** 
 * Creates new attributes for the chosen time granularities. Each
 * new attribute specifies as an event set the periods of time during
 * which a granularity exists. For example if the base time unit of
 * a time line was days and the chosen granularity was a weekend the
 * new attribute weekend-granularity would have each Saturday and Sunday
 * noted as an event. Additionally, each granularity can be numbered 
 * or unnumbered. That is, each weekend could be specified as an event
 * in the weekend-granularity attribute, or a new attribute can be 
 * created for each weekend; weekend-granularity-1, weekend-granularity-2,
 * and so on. Each of these attributes would have at most 1 event for 
 * each instance.
 *
 * Hopefully one day this class will be extended to work with time lines
 * that use explicit dates but for now it will work with time lines
 * that use an integer format where the user knows the implicit units
 * of time it represents.
 *
 * Valid filter-specific options are:<br>
 *
 * -N <br>
 * Specifies granularities will be numbered.
 *
 * -H <br>
 * Specify the half granularity. That is the time line for each instance
 * is divided in half. Each instance will contain 2 half granularities.
 *
 * -U <br>
 * Specify a user defined granularity where a number of units on the 
 * time line signifies the granularity. If the user chooses 10 then the
 * granularity is 10 units. The new attribute(s) will be named for this
 * value.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version $Revision: 0.9 $
 */

public class TimeGranularityFilter extends Filter 
  implements OptionHandler
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The string for Half granularity attribute name. */
  public static String HALF_GRANULARITY = "half-granularity";

  /** The string to append for user defined granularity
   *  attribute name. */
  public static String USER_GRANULARITY = "-granularity";

  /* Just so we can switch on granularity */
  private final static int HALF_INT = 1;
  private final static int USER_INT = 2;

  /** The delimiters to use for sets of events */
  public static char SETDELIMITER = '^';
  public static char SETBEGIN = '{';
  public static char SETEND = '}';

  /** For the numbered granularities we need to know the maximum
      time noted in the dataset. */
  private double maxTime;

  /** list of time sequence attributes to find events in */
  private Vector eventAttributeVector = null;

  /** numbered option */
  protected boolean numberedGranularities = false;

  /** create half granularity? */
  protected boolean halfGranularity = false;

  /** size of user defined granularity */
  protected int userGranularitySize = 0;

  /** names for the attributes of the user numbered granularities */
  private String [] numberedUserAttributeStringArray = null;

  /** Specifies debug info level 
   0: no debug info
   1: input to methods
   2: and output from methods
   3: and all sorts of stuff */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Sets the size of the user defined granularity in terms
   * of time line units.
   *
   * @param n
   * the number of units
   */

  public void setUserGranularitySize ( String n )
  {
    try
    {
      setUserGranularitySize ( Integer.parseInt ( n ) );
    }
    catch ( NumberFormatException ex )
    {
      setUserGranularitySize ( 10 );
    } // end else use default

  } // END public void setUserGranularitySize ( String n )

//********************************************************//

  /**
   * Sets the size of the user defined granularity in terms
   * of time line units.
   *
   * @param n
   * the number of units
   */

  public void setUserGranularitySize ( int n )
  {
    userGranularitySize = n;

  } // END public void setUserGranularitySize ( int n )

//********************************************************//

  /**
   * Returns the user granularity size
   *
   * @return
   * the number of units
   */

  public int getUserGranularitySize()
  {
    return ( userGranularitySize );

  } // END public int getUserGranularitySize()

//********************************************************//
  
  public String userGranularitySizeTipText()
  {
    return ( "Specify the number of time line units for the user defined granularity. A value of 0 or less indicates no user defined granularity." );
  } // END public String userGranularitySizeTipText()

//********************************************************//

  /**
   * Sets if the filter should create the half granularity.
   *
   * @param b
   * true to define half granularity, false otherwise
   */

  public void setHalfGranularity ( boolean b )
  {
    halfGranularity = b;

  } // END public void setHalfGranularity ( boolean b )

//********************************************************//

  /**
   * Returns if the filter should create the half granularity.
   *
   * @return
   * true to define half granularity, false otherwise
   */

  public boolean getHalfGranularity()
  {
    return ( halfGranularity );

  } // END public boolean getHalfGranularity()

//********************************************************//
  
  public String halfGranularityTipText()
  {
    return ( "Specify if the half granularity should be created." );
  } // END public String halfGranularityTipText()

//********************************************************//

  /**
   * Sets if the filter should create numbered granularities.
   *
   * @param b
   * true to create numbered granularities, false otherwise
   */

  public void setNumberedGranularities ( boolean b )
  {
    numberedGranularities = b;

  } // END public void setNumberedGranularities ( boolean b )

//********************************************************//

  /**
   * Returns if the filter should create numbered granularities.
   *
   * @return
   * true to define numbered granularities, false otherwise
   */

  public boolean getNumberedGranularities()
  {
    return ( numberedGranularities );

  } // END public boolean getNumberedGranularities()

//********************************************************//
  
  public String numberedGranularitiesTipText()
  {
    return ( "Specify if numbered granularities should be created." );
  } // END public String numberedGranularitiesTipText()

//********************************************************//

  /**
   * Finds the granularities for the instance specified.
   * If the numbered option is selected a string for each
   * value is returned.
   * @return
   * <ul>
   * <li>the granularity events</li>
   * <li>empty array if no events found</li>
   * </ul>
   */

  private String [] getGranularityEvents ( Instance instance, 
					   int granularityType )
  {
    // have to check each event attribute

    // all we have to do for the granularities defined so far is to find
    // the maximum ending time for an event for this instance

    double max = -1.0;
    double [] tempEventTime;
    String [] eventArray;

    int i, k;
    for ( i = 0; i < eventAttributeVector.size(); i++ )
    {      
      // make sure there's a value for this attribute
      if ( ! instance.isMissing
	   ( (Attribute)eventAttributeVector.elementAt ( i ) ) )
      {
	// get the value parsed into events
	eventArray = 
	  TSUtility.getEvents 
	  ( instance.stringValue
	    ( (Attribute)eventAttributeVector.elementAt ( i ) ) );

	for ( k = 0; k < eventArray.length; k++ )
	{
	  // get the values
	  tempEventTime = TSUtility.getValues ( eventArray [ k ] );

	  max = Math.max ( tempEventTime [ 1 ], max );

	} //end for each event 

      } // end if value is not missing

    } // end for each event attribute

    // ok, if we didn't find any event max should still be -1
    if ( max < 0 )
    {
      // just return an empty array
      return ( new String [ 0 ] );
      
    } // end if no events found    

    // the granularity events
    Vector granularityEventVector = new Vector();

    switch ( granularityType )
    {
    case HALF_INT:
      // take half the max and round up. 0 to this value is this first
      // event, this value to max is the second event
      granularityEventVector.add 
	( new StringBuffer ( "0" +
			     TSUtility.TIME_SEQUENCE_DELIMITER +
			     Math.ceil ( max / 2.0 ) ) );
      granularityEventVector.add 
	( new StringBuffer ( "" + ( Math.ceil ( max / 2.0 ) + 1 ) +
			     TSUtility.TIME_SEQUENCE_DELIMITER +
			     max ) );
      break;

    case USER_INT:
      // here we simply create events starting from 0 until we reach max
      for ( i = 0; i <= max; i += userGranularitySize )
      {
	// remember that since we start from zero we actually want ranges
	// from 0 to user size - 1
	granularityEventVector.add 
	  ( new StringBuffer ( "" + i +
			       TSUtility.TIME_SEQUENCE_DELIMITER +
			       ( i + userGranularitySize - 1 ) ) );
	
      } // end for each user granularity

      break;

    } // end switch granularity type
    
    String [] results;

    // ok, if the numbered option is selected we have to make
    // a bunch of new event sets.

    StringBuffer tempSB;
    
    if ( numberedGranularities )
    {
      results = new String [ granularityEventVector.size() ];

      for ( i = 0; i < results.length; i++ )
      {
	tempSB = (StringBuffer)granularityEventVector.elementAt ( i );

	tempSB.insert ( 0, SETBEGIN );
	tempSB.append ( SETEND );

	results [ i ] = tempSB.toString();
	
      } // end for each numbered event
      
    } // end if numbered option
    else
    {
      results = new String [ 1 ];

      tempSB = new StringBuffer();

      // set up the value
      tempSB.append ( SETBEGIN );

      // append the first event
      tempSB.append ( (StringBuffer)granularityEventVector.elementAt ( 0 ) );

      for ( i = 1; i < granularityEventVector.size(); i++ )
      {
	tempSB.append ( TSUtility.SET_DELIMITER );
	tempSB.append ( (StringBuffer)granularityEventVector.elementAt ( i ) );

      } // end for each unnumbered event

      tempSB.append ( SETEND );

      results [ 0 ] = tempSB.toString();

    } // end else not numbered

    return ( results );

  } // END private String getGranularityEvents ( Instance instance,
  // int granularityType )

//********************************************************//

  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input instance
   * structure (any values of instances contained in the object are 
   * ignored) but in addition to the structure, this filter does require 
   * to know the number of instances in the input data set.
   *
   * @return true if the outputFormat may be collected immediately
   * @exception Exception if the format couldn't be set successfully
   */

  public boolean setInputFormat ( Instances instanceInfo ) 
    throws Exception 
  {
    // use filter's set input format
    super.setInputFormat ( instanceInfo );

    // output format may NOT be collected right away
    // we have to set the values for our string attributes
    // first
    return ( false );

  } // END   public boolean setInputFormat ( Instances instanceInfo ) 
  //    throws Exception 

 //********************************************************//

  /**
   * Sets the output format. Takes the buffered input
   * instances and generates valid strings for the new
   * attributes.
   */
  
  private void setOutputFormat() 
  {
    // make a new instances for the output format
    Instances outputFormat = new Instances ( getInputFormat(), 0 );

    // did we find the event attributes yet?
    if ( eventAttributeVector == null )
    {
      findEventAttributes();

    } // end if event attribute vector not initialized

    // add half granularity?
    if ( halfGranularity )
    {
      // if numbered granularities is set we should make 2 attributes
      // one for the the first half, one for the second half
      if ( numberedGranularities )
      {
	// only add an attribute if one does not already exist
	if ( outputFormat.attribute ( HALF_GRANULARITY + "-1" ) == null )
	{
	  outputFormat.
	    insertAttributeAt ( new Attribute ( HALF_GRANULARITY + "-1", 
						null ),
				outputFormat.numAttributes() );
	} // end if attribute must be added

	// only add an attribute if one does not already exist
	if ( outputFormat.attribute ( HALF_GRANULARITY + "-2" ) == null )
	{
	  outputFormat.
	    insertAttributeAt ( new Attribute ( HALF_GRANULARITY + "-2", 
						null ),
				outputFormat.numAttributes() );
	} // end if attribute must be added

      } // end if numbered granularities
      else
      {
	// only add an attribute if one does not already exist
	if ( outputFormat.attribute ( HALF_GRANULARITY ) == null )
	{
	  outputFormat.
	    insertAttributeAt ( new Attribute ( HALF_GRANULARITY, null ),
				outputFormat.numAttributes() );
	
	} // end if attribute must be added

      } // end else not numbered
      
    } // end if half granularity
    
    // add user granularity?
    if ( userGranularitySize > 0 )
    {
      // if numbered granularities is set we should make a new
      // attribute for each
      if ( numberedGranularities )
      {
	// create the list of attribute names of it isn't there already
	if ( numberedUserAttributeStringArray == null )
	{
	  fillNumberedUserAttributeStringArray();

	} // end if we need to create the list of attribute names

	for ( int i = 0; i < numberedUserAttributeStringArray.length; i ++ )
	{
	  // only add if one does not already exist
	  if ( outputFormat.attribute 
	       ( numberedUserAttributeStringArray [ i ] ) == null )
	  {
	    outputFormat.
	      insertAttributeAt 
	      ( new Attribute ( numberedUserAttributeStringArray [ i ], null ),
		outputFormat.numAttributes() );

	  } // end if attribute must be added
	  
	} // end for each user granularity
	
      } // end if numbered granularities
      else
      {
	// only add if one does not already exist
	if ( outputFormat.attribute ( userGranularitySize +
				      USER_GRANULARITY ) == null )
	{
	  outputFormat.
	    insertAttributeAt ( new Attribute ( userGranularitySize +
						USER_GRANULARITY, null ),
				outputFormat.numAttributes() );
	  
	} // end if attribute must be added

      } // end else not numbered

    } // end if user granularity
 
    setOutputFormat ( outputFormat );
    
  } // END private void setOutputFormat() 

 //********************************************************//

  /**
   * Finds the event attributes in the input format. 
   */

  protected void findEventAttributes()
  {
    // make a new vector for storing the list
    // of event attributes
    eventAttributeVector = new Vector();

    // track is attribute is an event set
    boolean isEventSet;

    // have to find the max time
    maxTime = 0;
    String [] eventArray;
    double [] eventTimeArray;
    int i;

    // go through all the attributes in the input
    // checking the type for string
    for ( int a = 0; a < getInputFormat().numAttributes(); a++ )
    {
      if ( getInputFormat().attribute ( a ).isString() )
      {
	// let's be optimistic and say as far as we know this attribute
	// is an event set
	isEventSet = true;

	// check the values
	Enumeration enum = getInputFormat().attribute ( a ).
	  enumerateValues();

	// temp string for checking stuff
	String tempString;
	
	while ( enum.hasMoreElements() )
	{
	  tempString = (String)enum.nextElement();

	  if ( ! TSUtility.isTimeSequenceEventSet ( tempString) )
	  {
	    // I guess we were wrong to be optimistic
	    isEventSet = false;

	    // we don't have to look any longer at this attribute's
	    // values
	    break;

	  } // end if value is not an event attribute
	  else
	  {
	    eventArray = TSUtility.getEvents ( tempString );

	    for ( i = 0; i < eventArray.length; i++ )
	    {
	      eventTimeArray = TSUtility.getValues ( eventArray [ i ] );
	      maxTime = Math.max ( eventTimeArray [ 1 ], maxTime );

	    } // end for each event

	  } // end else remember max time

	} // end while enum has more

	// is this attribute an event set?
	if ( isEventSet )
	{
	  // add it to our list of event attributes then
	  eventAttributeVector.add ( getInputFormat().attribute ( a ) );

	} // end if attribute an event set
	
      } // end if attribute type string
      
    } // end for each attribute
    
  } // END protected void findEventAttributes()
  
 //********************************************************//

  /**
   * Creates a list of attribute names for the user specified
   * granularity when the numbered granularities option is set.
   */

  private void fillNumberedUserAttributeStringArray()
  { 
    Vector tempVector = new Vector();
    int i;

    for ( i = 0; i < maxTime; i += userGranularitySize )
    {
      tempVector.add ( userGranularitySize + 
		       USER_GRANULARITY + "-" + i + "-" + 
		       ( i + userGranularitySize - 1 ) );

    } // end for each user granularity

    numberedUserAttributeStringArray = new String [ tempVector.size() ];

    for ( i = 0; i < tempVector.size(); i++ )
    {
      numberedUserAttributeStringArray [ i ] = 
	(String)tempVector.elementAt ( i );

    } // end for each user granularity

  } // END void private fillNumberedUserAttributeStringArray()
  
 //********************************************************//

  /**
   * Input an instance for filtering. Ordinarily the
   * instance is processed and made available for output
   * immediately. Some filters require all instances be
   * read before producing output.
   *
   * @param instance the input instance
   *
   * @return true if the filtered instance may now be
   * collected with output().
   *
   * @exception IllegalStateException
   * if no input format has been defined.
   * 
   */

  public boolean input ( Instance instance )
  {
    // make sure input format is set
    if ( getInputFormat() == null ) 
      {
	throw new IllegalStateException ( "No input instance format defined" );
      }

    // if this is a new batch clear the queue and 
    // signify that a batch has begun
    if ( m_NewBatch )
      {
	resetQueue();
	m_NewBatch = false;

      } // end if new batch

    // save the instance for processing later
    bufferInput ( instance );

    // the output cannot be collected yet
    return ( false );

  } // END public boolean input ( Instance instance )

//********************************************************//

  /**
   * Signifies that this batch of input to the filter is
   * finished. 
   *
   * @return true
   * if there are instances pending output.
   *
   * @exception IllegalStateException
   * if no input structure has been defined.
   */

  public boolean batchFinished() 
  {
    // make sure we have an input format
    if ( getInputFormat() == null ) 
    {
      throw new IllegalStateException ( "No input instance format defined" );
    }

    // if the output format is not defined then set it and
    // generate the output instance
    if ( ! isOutputFormatDefined() ) 
      {
	setOutputFormat();

	// need a temp string array for granularity events
	String [] tempStringArray;

	// a spare index for those spontaneous for loops
	int k;

	// go through all the buffered instances
	for ( int i = 0; i < getInputFormat().numInstances(); i++ )
	{	  	
	  // make a new instance with the output data set
	  Instance newInstance = new Instance ( getOutputFormat().
						numAttributes() );
	  newInstance.setDataset ( getOutputFormat() );
	  
	  // set all the attributes of this new instance
	  // with values from a buffered input instance
	  for ( int a = 0; a < getInputFormat().numAttributes(); a++ )
	  {
	    if ( getOutputFormat().attribute ( a ).isNumeric() )
	    {
	      newInstance.setValue ( a, getInputFormat().instance ( i ).
				     value ( a ) );

	    }  // end if numeric set value using value
	    else
	    {
	      newInstance.setValue ( a, getInputFormat().instance ( i ).
				     stringValue ( a ) );
	    } // end else set using stringValue

	  } // end for all input attributes

	  // should we create the half granularity
	  if ( halfGranularity )
	  {
	    tempStringArray = 
	      getGranularityEvents ( getInputFormat().instance ( i ),
				     HALF_INT );

	    // if numbered granularity is set have to put each value in
	    // its own attribute
	    if ( numberedGranularities )
	    {
	      newInstance.setValue ( getOutputFormat().
				     attribute ( HALF_GRANULARITY + "-1" ), 
				     tempStringArray [ 0 ] );

	      newInstance.setValue ( getOutputFormat().
				     attribute ( HALF_GRANULARITY + "-2" ),
				     tempStringArray [ 1 ] );
	    } // end if numbered
	    else
	    {
	      newInstance.setValue ( getOutputFormat().
				     attribute ( HALF_GRANULARITY ),
				     tempStringArray [ 0 ] );

	    } // end else not numbered

	  } // end if half granularity
	  
	  // should we create the user granularity
	  if ( userGranularitySize > 0 )
	  {
	    tempStringArray = getGranularityEvents 
	      ( getInputFormat().instance ( i ),
		USER_INT );
	    
	    // if numbered granularity is set have to put each value in
	    // its own attribute
	    if ( numberedGranularities )
	    {
	      for ( k = 0; k < tempStringArray.length; k++ )
	      {
		newInstance.setValue 
		  ( getOutputFormat().
		    attribute ( numberedUserAttributeStringArray [ k ] ),
		    tempStringArray [ k ] );
		
	      } // end for each granularity

	    } // end if numbered
	    else
	    {
	      // have to make sure there is at least one element in the array
	      // of course
	      if ( tempStringArray.length > 0 )
	      {
		newInstance.setValue ( getOutputFormat().
				       attribute ( userGranularitySize + 
						   USER_GRANULARITY ),
				       tempStringArray [ 0 ] );

	      } // end if there's a value to put
	      else
	      {
		newInstance.setValue ( getOutputFormat().
				       attribute ( userGranularitySize + 
						   USER_GRANULARITY ),
				       "" );
	      } // end else put blank

	    } // end else not numbered

	  } // end if user granularity

	  // all set, push her on out
	  push ( newInstance );
	  
	} // end for all input instances
	
      } // end if output format not defined
    
    flushInput();
    m_NewBatch = true;
    return ( numPendingOutput() != 0 );

  } // END public boolean batchFinished() 

//********************************************************//

  /**
   * Returns a string describing this filter.
   *
   * @return
   * a description of the filter suitable for displaying
   * in the explorer/experimenter gui
   *
   */

  public String globalInfo()
  {
    return ( "Creates event attributes for different time granularities." );

  } // END public String globalInfo()

//********************************************************//

  /**
   * Returns an enumeration describing the available options
   *
   * @return
   * an enumeration of all the available options
   */

  public Enumeration listOptions() 
  {
    Vector newVector = new Vector ( 6 );

    newVector.addElement ( new Option 
			   ( "\tCreates a new attribute for each granule possible. This allows for specific periods of time along the time line to be identified. If this option is not selected a single attribute will be created to hold all the granules.",
			     "N", 0, "-N" ) );

    newVector.addElement ( new Option 
			   ( "\tCreates a half granularity where each instance is dived into a first have and second half of the time line. this will be changed later to accept the percentage of time to split a time line into rather than it being 50% always.",
			     "H", 1, "-H <value>" ) );

    newVector.addElement ( new Option 
			   ( "\tSpecifies the number of time units to include in each granule. For example, if the time line was in seconds and you want a minute granularity set this option to 60.",
			     "U", 1, "-U <value>" ) );

    return ( newVector.elements() );

  } // END public Enumeration listOptions() 

//********************************************************//

  /**
   * Parses a list of options for this object. Valid
   * options are:
   *
   * @param options
   * the list of options as an array of strings
   *
   * - attributes <br>
   * Specify the attributes to find events in
   *
   * -N <br>
   * Specifies granularities will be numbered.
   *
   * -H <br>
   * Specify the half granularity. That is the time line for each instance
   * is divided in half. Each instance will contain 2 half granularities.
   *
   * -U <br>
   * Specify a user defined granularity where a number of units on the 
   * time line signifies the granularity. If the user chooses 10 then the
   * granularity is 10 units. The new attribute(s) will be named for this
   * value.
   * 
   * @exception Exception
   * if an option is not supported
   */

  public void setOptions ( String[] options ) throws Exception 
  {

    setUserGranularitySize ( Utils.getOption ( 'U', options ) );

    setNumberedGranularities ( Utils.getFlag ( 'N', options ) );
    setHalfGranularity ( Utils.getFlag ( 'H', options ) );
    
  } // END public void setOptions ( String[] options ) 
  // throws Exception

//********************************************************//

  /**
   * Gets the current settings of the filter.
   *
   * @return
   * an array of strings suitable for passing to setOptions
   */

  public String [] getOptions() 
  {
    String [] options = new String [6];
    int current = 0;

    if ( halfGranularity )
    {
      options [ current++ ] = "-H"; options [ current++ ] = "";
    }

    if ( numberedGranularities )
    {
      options [ current++ ] = "-N"; options [ current++ ] = "";
    }

    if ( userGranularitySize > 0 )
    {
      options [ current++ ] = "-U"; options [ current++ ] =
				      "" + getUserGranularitySize();
    }

    // fill anything left over with empties
    while ( current < options.length ) 
    {
      options [ current++ ] = "";
    }
    
    // all done cause we're good like that
    return ( options );

  } // END public String [] getOptions()

//********************************************************//
  
  /**
   * Main method for testing this class.
   *
   * @param argv
   * should contain arguments to the filter: 
   * use -h for help
   */

  public static void main ( String [] argv ) 
  {
    try 
      {
	if ( Utils.getFlag ( 'b', argv ) ) 
	  {
	    Filter.batchFilterFile ( new TimeGranularityFilter(), argv ); 
	  } 
	else 
	  {
	    Filter.filterFile ( new TimeGranularityFilter(), argv );
	  }
      } 
    catch ( Exception ex ) 
      {
	System.out.println ( ex.getMessage() );
      }

  } // END public static void main ( String [] argv )

//********************************************************//

} // END public class TimeGranularityFilter extends Filter 
  //  implements OptionHandler
