package wpi.filters;

import java.io.*;
import java.util.*;

import wpi.timesequence.*;

import weka.core.*;
import weka.filters.*;

/** 
 * Assumes intances form a single instance containing time sequence 
 * attributes. Each instance in instances comprises a point in time. 
 * The values for each instance corresponding to a single point in
 * time are concatanated to form a single value for each attribute
 * using a delimeter ":" to seperate each value inside a
 * time sequence attribute. A time_number attribute is added for
 * relative reference to points in time.
 * <br>
 * Please note: if the delimeter appears in the values of the data
 * being filtered it will be replaced by a "-"
 *
 * Valid filter-specific options are:<p>
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version $Revision: 0.9 $
 */

public class TimeSequenceInstanceFilter extends Filter 
  implements OptionHandler
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The character used to replace occurrances of the
      delimiter in the actual data being filtered */
  static public char DELIMITER_REPLACE = '-';

  /** The name of the time line attribute that this 
      filter creates */
  static public String TIMELINE = "time-line";

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Sets the format of the input instances.
   *
   * @param instanceInfo an Instances object containing the input instance
   * structure (any values of instances contained in the object are 
   * ignored) but in addition to the structure, this filter does require 
   * to know the number of instances in the input data set.
   *
   * @return true if the outputFormat may be collected immediately
   * @exception Exception if the format couldn't be set successfully
   */

  public boolean setInputFormat ( Instances instanceInfo ) 
    throws Exception 
  {
    // use filter's set input format
    super.setInputFormat ( instanceInfo );
    
    // the output format may not be collected right away
    return ( false );

  } // END public boolean setInputFormat ( 
  // Instances instanceInfo ) throws Exception 

//********************************************************//

  /**
   * Input an instance for filtering. Filter requires
   * all instances be read before producing output.
   *
   * @param instance the input instance
   *
   * @return true if the filtered instance may now be
   * collected with output().
   *
   * @exception IllegalStateException if no input format has been defined.
   * 
   */

  public boolean input ( Instance instance )
  {
    // make sure input format is set
    if ( getInputFormat() == null ) 
      {
	throw new IllegalStateException( "No input instance format defined" );
      }

    // if this is a new batch clear the queue and 
    // signify that a batch has begun
    if ( m_NewBatch )
      {
	resetQueue();
	m_NewBatch = false;

      } // end if new batch

    // save the instance for processing later
    bufferInput ( instance );

    // the output cannot be collected yet
    return ( false );
    
  } // END public boolean input ( Instance instance )

//********************************************************//

  /**
   * Signifies that this batch of input to the filter is
   * finished. 
   *
   * @return true
   * if there are instances pending output.
   *
   * @exception IllegalStateException
   * if no input structure has been defined.
   */

  public boolean batchFinished() 
  {
    // make sure we have an input format
    if ( getInputFormat() == null ) 
      {
	throw new IllegalStateException ( "No input instance format defined" );
      }
    
    // if the ouput format is not defined then set it and
    // generate the output instance
    if ( ! isOutputFormatDefined() ) 
      {
	setOutputFormat();
	
	String [] newValues = new String [ getInputFormat().
					 numAttributes() + 1 ];
	
	// construct the time line attribute values
	// it should go first in the array
	newValues[0] = "0";
	for ( int i = 1; i < getInputFormat().numInstances(); i++ )
	  {
	    newValues[0] = newValues[0].concat 
	      ( "" + TSUtility.TIME_SEQUENCE_DELIMITER + i );
	  }
	
	// construct new values from pending input instances
	
	// start by initializing each string value with the first
	// instance
	for ( int a = 0; a < getInputFormat().numAttributes(); a++ )
	  {
	    // remember to filter each value for the delimiter
	    // and replace it with the delimiter replacement
	    newValues[a+1] = ((Instance)(getInputFormat().
					 instance ( 0 ))).toString ( a ).
	      replace ( TSUtility.TIME_SEQUENCE_DELIMITER, DELIMITER_REPLACE );

	  } // end for each attribute of the first instance
	
	// now append the value for all the other instances using
	// the delimiter of course
	for ( int i = 1; i < getInputFormat().numInstances(); i++ )
	  {
	    for ( int a = 0; a < getInputFormat().numAttributes(); a++ )
	      {
		// remember to filter each value for the delimiter
		// and replace it with the delimiter replacement
		newValues[a+1] = newValues[a+1].
		  concat ( "" + TSUtility.TIME_SEQUENCE_DELIMITER + 
			   ((Instance)(getInputFormat().instance ( i ))).
			   toString ( a ).
			   replace ( TSUtility.TIME_SEQUENCE_DELIMITER, 
				     DELIMITER_REPLACE ) );

	      } // end for each attribute
	    
	  } // end for each instance
	
	// make a new instance with the output data set
	Instance newInstance = new Instance ( getOutputFormat().
					      numAttributes() );
	newInstance.setDataset ( getOutputFormat() );
	
	// set the values
	for ( int a = 0; a < newInstance.numAttributes(); a++ )
	  {
	    newInstance.setValue ( a, newValues[a] );
	    
	  } // end for setting each attribute value
	
	// all done, send it on its way
	push ( newInstance );

      } // end if output format is not defined

    flushInput();
    m_NewBatch = true;
    return ( numPendingOutput() != 0 );

  } // END public boolean batchFinished() 

//********************************************************//

  /**
   * Sets the output format. Takes the buffered input
   * instances and generates valid strings for the new
   * attributes.
   */
  
  private void setOutputFormat() 
  {
    Instances newData;
    FastVector newAtts, newVals;
      
    // make new attributes, need an extra for time line
    newAtts = new FastVector ( getInputFormat().numAttributes() + 1 );

    newAtts.addElement ( new Attribute ( TIMELINE, null ) );

    // now add string attributes with the same names as the
    // input attributes
    for (int j = 0; j < getInputFormat().numAttributes(); j++) 
      {
	Attribute att = getInputFormat().attribute(j);

	// make all the new attributes strings
	// the null tells the constructor it is a string 
	// attribute
	newAtts.addElement ( new Attribute ( att.name(), null ) );
	
      }
      
    // Construct new header
    newData = new Instances ( getInputFormat().relationName(), newAtts, 0 );

    setOutputFormat ( newData );

  } // END private void setOutputFormat() 
  
//********************************************************//

  /**
   * Returns a string decribing this filter.
   *
   * @return
   * a description of the filter suitable for displaying
   * in the explorer/experimenter gui
   *
   */

  public String globalInfo()
  {
    return ( "Assumes intances form a single instance " +
	     "containing time sequence attributes. Each " + 
	     "instance in instances comprises a point in " +
	     "time. The values for each instance " +
	     "corresponding to a single point in time are " +
	     "concatanated to form a single value for each " +
	     "attribute using a delimeter \"" + 
	     TSUtility.TIME_SEQUENCE_DELIMITER + "\" to " +
	     "seperate each value inside a time sequence " +
	     "attribute. A time_number attribute is added " +
	     "for relative reference to points in time. " +
	     "Please note: if the delimeter appears in the " +
	     "values of the data being filtered it will be " +
	     "replaced by a \"-\"" );

  } // END public String globalInfo()

//********************************************************//

  /**
   * Returns an enumeration describing the available options
   *
   * @return an enumeration of all the available options
   */

  public Enumeration listOptions() 
  {
    Vector newVector = new Vector(0);

    // no options yet

    return newVector.elements();

  } // END public Enumeration listOptions() 

//********************************************************//

  /**
   * Parses a list of options for this object. Valid options are:<p>
   *
   * @param options
   * the list of options as an array of strings
   *
   * @exception Exception
   * if an option is not supported
   */

  public void setOptions ( String[] options ) throws Exception 
  {
    // no options yet

  } // END   public void setOptions ( String[] options ) 
  // throws Exception

//********************************************************//

  /**
   * Gets the current settings of the filter.
   *
   * @return
   * an array of strings suitable for passing to setOptions
   */

  public String [] getOptions() 
  {
    String [] options = new String [0];
    int current = 0;

    // no options yet

    return options;

  } // END public String [] getOptions()

//********************************************************//
  
  /**
   * Main method for testing this class.
   *
   * @param argv
   * should contain arguments to the filter: 
   * use -h for help
   */

  public static void main ( String [] argv ) 
  {
    try 
      {
	if ( Utils.getFlag ( 'b', argv ) ) 
	  {
	    Filter.batchFilterFile ( new TimeSequenceInstanceFilter(), argv ); 
	  } 
	else 
	  {
	    Filter.filterFile ( new TimeSequenceInstanceFilter(), argv );
	  }
      } 
    catch ( Exception ex ) 
      {
	System.out.println ( ex.getMessage() );
      }

  } // END public static void main ( String [] argv )

//********************************************************//

} // END public class TimeSequenceInstanceFilter extends Filter 
// implements OptionHandler


