package wpi.timesequence;

import java.io.*;
import java.util.*;

import weka.filters.*;
import weka.core.*;

/*
  Maintainance log started on February 24, 2003

  February 24, 2003 - made consistent the convention of using a lower case
		      letter to start a method name. -kap
*/

/** 
 * Appends instances to a target data set. Any attributes
 * existing in the append data set that do not appear
 * in the target data set are added. All instances in the 
 * target are set to unknown for those attributes.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version $Revision: 0.9 $
 */

public class AppendInstances implements OptionHandler
{
  
//********************************************************//
// ----- Data Members -----
//********************************************************//  
  
  /** Name of the target file used */
  private String targetName;
  
  /** Name of the append file last used */
  private String appendName;

  /** Target Instances **/
  private Instances target;

  /** Append Instances **/
  private Instances append;

//********************************************************//
// ----- Methods-----
//********************************************************//
  
  /**
   * Constructor.
   */
  
  public AppendInstances()
  {
    init();
    
  } // END public AppendInstances()

//********************************************************//

  /**
   * Constructor - uses options to initialize
   *
   * @param options
   * the list of options as an array of strings
   *
   * @exception Exception
   * if an option is not supported
   */

  public AppendInstances ( String[] options ) throws Exception 
  {
    init();
    setOptions ( options );

  } // END public AppendInstances ( String[] options )

//********************************************************//

  /**
   * Initializes data members
   */

  private void init()
  {
    // set file names to null
    targetName = null;
    appendName = null;
    target = null;
    append = null;

  } // END private void init()

//********************************************************//

  /**
   * Sets the target instances to use.
   *
   * @param name
   * name of the data file containing the instances
   *
   * @return
   * <code>true</code> if all set to append instances
   */

  public boolean setTarget ( String name ) throws Exception 
  {
    targetName = name;

    // try loading the data set
    
    BufferedReader r = new BufferedReader ( new FileReader ( targetName ) );

    return ( setTarget ( new Instances ( r ) ) );

  } // END public boolean setTarget ( String name )

//********************************************************//

  /**
   * Set the target instances
   *
   * @param instances
   * instances to use as target
   *
   * @return
   * <code>true</code> if all set to append instances
   */

  public boolean setTarget ( Instances instances )
  {
    target = instances;

    return ( ready() );

  } // END public boolean setTarget ( Instances instances )

//********************************************************//

  /**
   * Appends instances to the target.
   *
   * @param name
   * name of the data file containing the instances to append
   *
   * @return
   * <code>true</code> if instances were appended successfully
   */

  public boolean append ( String name ) throws Exception 
  {
    // try loading the data set
    
    BufferedReader r = new BufferedReader ( new FileReader ( name ) );
    
    // try appending it, if it works record the name
    if ( append ( new Instances ( r ) ) )
    {
      appendName = name;
      return ( true );
    }
    else
    {
      return ( false );
    }
    
  } // END public boolean append ( String name )

//********************************************************//

  /**
   * Appends instances to the target.
   *
   * @param instances
   * instances to append
   *
   * @return
   * <code>true</code> if instances were appended successfully
   */

  public boolean append ( Instances instances ) throws Exception
  {
    // do we have a real set of instances and are we ready?
    if ( (instances == null) || (! ready()) )
    {
      return ( false );
    }
    
    // ok, let's fry this puppy
    append = new Instances ( instances );
    
    // we may be able to skip a bunch of work if the header
    // of this instances matches that of target
    // try simply adding the instances
    if ( add ( append ) )
    {
      return ( true );
    }
    
    // just have to do things brute force
    Instances temp = appendInstances ( target, append );
    
    // did it seem to work?
    if ( temp != null )
    {
      // set target
      target = temp;
      
      // set append back to the original instances data set
      // so it makes sense if people ask for the last instances
      // that were added successfully
      append = instances;
      
      return ( true );
      
    } // end if it looks like it worked
    
    // guess it didn't work
    return ( false );
    
  } // END public boolen append ( Instances instances )

//********************************************************//

  /**
   * Returns a new data set that has all the attributes
   * from both data sets and all data from both data sets.
   * Attributes apearing in both input data sets occur only
   * once in the returned data set. When attributes with
   * the same name are found to differ they are combined
   * as follows:
   * <br>
   * If both are nominal the valid values lists are combined.
   * <br>
   * If their types differ a string attribute with the same
   * name is created.
   * <br>
   * Instance values are copied from the original data sets
   * and transformed if necessary based on the return data
   * set's attribute types.
   * <br>
   * The new data set retains the relation name of the first
   * or target data set.
   *
   * @param i1
   * data set to append instances on to
   *
   * @param i2
   * data set to be appended
   *
   * @return
   * the new data set
   */

  static public Instances appendInstances ( Instances i1, 
					    Instances i2 )
    throws Exception
  {
    // I think I have to rewrite this

    // make three lists of attributes
    // the atts from i1, the atts from i2 and an empty list
    // (may need more lists to track which atts came from which instances)

    // attributes added to i1
    FastVector i1NewAtts = new FastVector ( i2.numAttributes() );

    // attributes added to i2
    FastVector i2NewAtts = new FastVector ( i1.numAttributes() );

    // attributes common to both i1 and i2
    FastVector newAtts = new FastVector ( i1.numAttributes() +
					  i2.numAttributes() );

    // attributes with the same name in i1 and i2 but
    // differ in type or valid values
    FastVector differentAtts = new FastVector ();

    // add to the lists appropriately
    for ( int i1n = 0; i1n < i1.numAttributes(); i1n++ )
    {
      // get a reference to i2's attribute with same name
      Attribute i2att = i2.attribute ( i1.attribute ( i1n ).name() );
      
      // is it in i2?
      if ( i2att != null )
      {
	// is it the same as i1's?
	if ( i1.attribute ( i1n ).equals ( i2att ) )
	{
	  newAtts.addElement ( i2att );
	  
	} // end if atts the same
	else
	{
	  // the potential values for this attribute
	  FastVector values = null;
	  
	  // if both are nominal just make a new nominal att
	  // with all the values from i1 and i2
	  if ( i1.attribute ( i1n ).isNominal() && i2att.isNominal() )
	  {
	    // get a copy of the i1 att values
	    values = new FastVector ( i1.attribute ( i1n ).
				      numValues() +
				      i2att.numValues() );
	    
	    // add the i1att values
	    Enumeration ev = i1.attribute ( i1n ).enumerateValues();
	    while ( ev.hasMoreElements() )
	    {
	      values.addElement ( ev.nextElement() );
	      
	    } // end while more values
	    
	    // add the i2att values
	    ev = i2att.enumerateValues();
	    while ( ev.hasMoreElements() )
	    {
	      Object o = ev.nextElement();
	      // check if value is not in the list add it
	      if ( values.indexOf ( o ) < 0 )
	      {
		values.addElement ( o );
		
	      } // end if value not in list yet
	      
	    } // end while more values
	    
	  } // end if both are nominal
	  
	  // ok, make a new attribute and add it to the
	  // different list
	  differentAtts.addElement ( new Attribute ( i2att.name(),
						     values ) );
	  // Note: if the atts were not both nominal
	  // a string attribute is created
	  // this is specified by values == null
	  
	} // end else have to merge attributes
	
      } // end if atts with same name in i2
      else
      {
	i2NewAtts.addElement ( i1.attribute ( i1n ) );
	
      } // end else add att to new i2 list
      
    } // end for each i1 attribute
    
    // now we add attributes from i2
    // this is a little different since all we have to do is
    // make sure that we did not already add the attributes
    // to the new and different atts lists.
    for ( int i2n = 0; i2n < i2.numAttributes(); i2n++ )
    {
      // first check to see if we took care of this attribute
      // already
      
      // is it new Atts?
      if ( newAtts.indexOf ( ( i2.attribute ( i2n ) ) ) < 0 )
      {
	// is it in different Atts?
	boolean found = false;
	
	Enumeration atts = differentAtts.elements();
	while ( atts.hasMoreElements() )
	{
	  if ( i2.attribute ( i2n ).name().equals 
	       ( ((Attribute)atts.nextElement()).name()) )
	  {
	    // we found it!
	    found = true;
	    // no need to continue in the while loop
	    break;
	  } // end if the names are the same
	  
	} // end while more atts
	
	// did we find it? if not add it to i1 new atts
	if ( ! found )
	{
	  i1NewAtts.addElement ( i2.attribute ( i2n ) );
	  
	} // end if att not found
	
      } // end if not in new Atts
      
    } // end for each i2 attribute

    // make a new Instances with this set of atts

    // ok, now we can append the attributes from 
    // i2 new atts, i1 new atts and different atts to new atts
    newAtts.appendElements ( i2NewAtts );
    newAtts.appendElements ( i1NewAtts );
    newAtts.appendElements ( differentAtts );

    // make a new Instances

    // use only i1's relation name since we are appending to that
    // data set
    Instances outputInstances = new Instances ( i1.relationName(),
						newAtts,
						i1.numInstances() +
						i2.numInstances() );

    // add each new instance to the new instances
    for ( int i = 0; i < i1.numInstances(); i++ )
    {
      // make a new instance with the num of attributes in out output
      Instance instance = new Instance ( outputInstances.numAttributes() );
      
      // set the data set to the output instances
      instance.setDataset ( outputInstances );
      
      // for each attribute in i1 get the value and set the corresponding
      // attribute in the new instance
      for ( int i1att = 0; i1att < i1.numAttributes(); i1att++ )
      {
	// how we set the value depends on the attribute type
	// in the new instance
	
	// get the attribute with the same name as the i1 instance
	Attribute att = outputInstances.
	  attribute ( i1.attribute ( i1att ).name() );
	
	switch ( att.type() )
	{
	case Attribute.NUMERIC:
	  instance.setValue ( att, i1.instance ( i ).value ( i1att ) );
	  break;
	  
	case Attribute.NOMINAL:
	case Attribute.STRING:
	  // have to check if the i1 instance is numeric
	  if ( i1.attribute ( i1att ).isNumeric() )
	  {
	    instance.setValue ( att, 
				Double.toString ( i1.instance ( i ).
						  value ( i1att ) ) );
	  }
	  else
	  {
	    instance.setValue ( att, 
				i1.instance ( i ).
				stringValue ( i1att ) );
	  }
	  break;
	  
	default:
	  throw new Error ( "Unknown attribute type in appendInstances" );
	  
	} // end switch on att type
	
      } // end for each att in i1
      
	// add the new instance to the out put
      outputInstances.add ( instance );
      
    } // end for each i1 instance
    
    // do the same for i2
    for ( int i = 0; i < i2.numInstances(); i++ )
    {
      // make a new instance with the num of attributes in out output
      Instance instance = new Instance ( outputInstances.numAttributes() );
      
      // set the data set to the output instances
      instance.setDataset ( outputInstances );
      
      // for each attribute in i1 get the value and set the corresponding
      // attribute in the new instance
      for ( int i2att = 0; i2att < i2.numAttributes(); i2att++ )
      {
	// how we set the value depends on the attribute type
	// in the new instance
	
	// get the attribute with the same name as the i1 instance
	Attribute att = outputInstances.
	  attribute ( i2.attribute ( i2att ).name() );
	
	switch ( att.type() )
	{
	case Attribute.NUMERIC:
	  instance.setValue ( att, i2.instance ( i ).value ( i2att ) );
	  break;
	  
	case Attribute.NOMINAL:
	case Attribute.STRING:
	  // have to check if the i2 instance is numeric
	  if ( i2.attribute ( i2att ).isNumeric() )
	  {
	    instance.setValue ( att, 
				Double.toString ( i2.instance ( i ).
						  value ( i2att ) ) );
	  }
	  else
	  {
	    instance.setValue ( att, 
				i2.instance ( i ).
				stringValue ( i2att ) );
	  }
	  break;
	  
	default:
	  throw new Error ( "Unknown attribute type in appendInstances" );
	  
	} // end switch on att type
	
      } // end for each att in i2
      
	// add the new instance to the out put
      outputInstances.add ( instance );
      
    } // end for each i1 instance
    
    // done :)
    return ( outputInstances );
    
  } // END static public Instances appendInstances ( Instances instances1, 
  //			       Instances instances2 )
  
//********************************************************//

  /**
   * Adds the given instances to target.
   */

  private boolean add ( Instances instances )
  {
    // is it possible?
    if ( ! target.equalHeaders ( instances ) )
    { 
      return ( false );
    }
    
    Enumeration inst = instances.enumerateInstances();

    while ( inst.hasMoreElements() )
    {
      target.add ( (Instance)inst.nextElement() );
    }
    
    return ( true );

  } // END private boolean add ( Instances instances )

//********************************************************//

  /**
   * Returns the target file name.
   *
   * @return
   * name of the file
   */

  public String getTargetName()
  {
    return ( targetName );

  } // END public String getTargetName()

//********************************************************//

  /**
   * Returns the target instances.
   *
   * @return
   * the target instances
   */

  public Instances getTarget()
  {
    return ( target );

  } // END public Instances getTarget()

//********************************************************//

  /**
   * Returns the file name of the last data set appended.
   *
   * @return
   * name of the file
   */

  public String getAppendName()
  {
    return ( appendName );

  } // END public String getAppendName()

//********************************************************//

  /**
   * Returns the last instances appended.
   *
   * @return
   * the last appended instances
   */

  public Instances getAppend()
  {
    return ( append );

  } // END public Instances getAppend()

//********************************************************//

  /** 
   * Checks if we are ready to appendinstances to the target.
   */

  public boolean ready()
  {
    if ( target != null )
    {
      return ( true );
    }
    
    return ( false );

  } // END public boolean ready()

//********************************************************//
  
  /**
   * Returns a string decribing this filter.
   *
   * @return
   * a description of the filter suitable for displaying
   * in the explorer/experimenter gui
   *
   */

  public String globalInfo()
  {
    return ( "coming soon" );

  } // END public String globalInfo()

//********************************************************//

  /**
   * Returns an enumeration describing the available options
   *
   * @return an enumeration of all the available options
   */

  public Enumeration listOptions() 
  {
    Vector newVector = new Vector ( 2 );
    
    newVector.
      addElement ( new Option 
		   ( "\tThe target data file.\n" +
		     "\tInstances will be appended to this file.",
		     "t", 1, "-t <value>" ) );
    
    newVector.
      addElement ( new Option 
		   ( "\tThe data file containing instances to append.\n",
		     "a", 1, "-a <value>" ) );
    
    return newVector.elements();

  } // END public Enumeration listOptions() 

//********************************************************//

  /**
   * Parses a list of options for this object. Valid options are:<p>
   *
   * @param options
   * the list of options as an array of strings
   *
   * @exception Exception
   * if an option is not supported
   */

  public void setOptions ( String[] options ) throws Exception 
  {

    setTarget ( Utils.getOption ( 't', options ) );
    append ( Utils.getOption ( 'a', options ) );

  } // END   public void setOptions ( String[] options ) 
  // throws Exception

//********************************************************//

  /**
   * Gets the current settings of the filter.
   *
   * @return
   * an array of strings suitable for passing to setOptions
   */

  public String [] getOptions() 
  {
    String [] options = new String [4];
    int current = 0;
    
    options[current++] = "-t"; options[current++] = getTargetName();
    options[current++] = "-a"; options[current++] = getAppendName();

    // fill anything left over with empty's
    while ( current < options.length ) 
    {
      options[current++] = "";
    }
    
    return ( options );


  } // END public String [] getOptions()

//********************************************************//

  /**
   * Main method for testing this class. Result is sent
   * to standard out.
   *
   * @param argv should contain arguments.
   */

  public static void main ( String [] argv ) 
  {
    AppendInstances appendInstances = null;

    try
    {
      // make a new one with options
      appendInstances = new AppendInstances ( argv );
      
    }
    catch ( Exception ex ) 
    {
      System.err.println ( ex.getMessage() );
      System.out.println ( "\nAppendInstances -t <file> -a <file>" );
      System.out.println ( "-t <file> " +
			   "The target data file. " +
			   "Instances will be appended to this file." );			   
      System.out.println ( "-a <file> " +
			   "The data file containing instances to append." );
      System.out.println ( "Result is sent to standard out." );
      return;
    }
    
    // did it work?
    if ( appendInstances != null )
    {
      System.out.println ( appendInstances.getTarget() );
    }
    
  } // END public static void main ( String [] argv )

//********************************************************//

} // END public class AppendInstances
