package wpi.timesequence;

import java.io.*;
import java.util.*;

import weka.core.*;

/*
  Maintainance log started on February 24, 2003

  February 24, 2003 - made consistent the convention of using a lower case
		      letter to start a method name. -kap

  February 22, 2004 - added getNominalValues method to parse symbolic
                      (non-numeric) sequences into seperate values. -kap
*/

/** 
 * This class contains methods for manipulating
 * time sequence attributes and events.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version $Revision: 0.9 $
 */

public class TSUtility
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The delimiters to use for time sequence values */
  public static char TIME_SEQUENCE_DELIMITER = ':';

  /** The delimiters to use for sets of events */
  public static char SET_DELIMITER = '^';
  public static char SET_BEGIN = '{';
  public static char SET_END = '}';

  /** The name of the data set attribute representing the time
      line for an instance. */
  public static String TIME_LINE_NAME = "time-line";

  /** Distance at which two points in time can be considered
   * the same. The value is absolute since a percentage
   * requires knowing the total length of the time scale.
   * The default value assumes a time unit of seconds. */
  public static double DISTANCE_TOLERANCE = 1.0;

  /** Specifies debug info level 
   0: no debug info
   1: input to methods
   2: and output from methods
   3: and all sorts of stuff */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Parses a time sequence attribute's value into
   * individual nominal values. 
   * Missing values are mapped to ?.
   *
   * @param v
   * the instance from which the value will come
   *
   * @return
   * an array of values as strings
   */

  public static String [] getNominalValues ( String v )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.print ( "getNominalValues ( " + v + " ) called;" );
    } // end switch debug

    // we could check if the value is a time sequence here
    // but let's assume it is... at worst we simply pass
    // back a list of 1 value
    StringTokenizer stringTokenizer = 
      new StringTokenizer ( v,
			    "" + TIME_SEQUENCE_DELIMITER );
    
    Vector value = new Vector();
    
    while ( stringTokenizer.hasMoreTokens() )
    {
      String temp = stringTokenizer.nextToken();

      switch ( debug ) {
      case 3:
	System.out.print ( " token : " + temp );      
      } // end switch debug

      // have to make sure there is a real value there
      // we will accept unknowns ( _ ) as valid and replace
      // them with ? for the purposes of this comparison
      if ( temp.equals ( "_" ) )
      {
	value.addElement ( "?" );
      }
      else
      {
	value.addElement ( temp );
	
      } // end else
      
    } // end while more tokens
    
    // make an array to return just the double values
    String [] dArray = new String [ value.size() ];

    for ( int i = 0; i < value.size(); i++ )
    {
      dArray [ i ] = (String)value.elementAt ( i );

    } // end for each values

    // all done
    return ( dArray );

  } // END public static String [] getNominalValues ( String v )

//********************************************************//

  /**
   * Parses a time sequence attribute's value into
   * individual values. Non-numeric are mapped to NaN.
   * Missing values are mapped to 0.
   *
   * @param v
   * the instance from which the value will come
   *
   * @return
   * an array of values
   */

  public static double [] getValues ( String v )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.print ( "getValues ( " + v + " ) called;" );
    } // end switch debug

    // we could check if the value is a time sequence here
    // but let's assume it is... at worst we simply pass
    // back a list of 1 value
    StringTokenizer stringTokenizer = 
      new StringTokenizer ( v,
			    "" + TIME_SEQUENCE_DELIMITER );
    
    Vector value = new Vector();
    
    while ( stringTokenizer.hasMoreTokens() )
    {
      String temp = stringTokenizer.nextToken();

      switch ( debug ) {
      case 3:
	System.out.print ( " token : " + temp );      
      } // end switch debug

      // have to make sure there is a real value there
      // we will accept unknowns ( _ and ? ) as valid and replace
      // them with 0 for the purposes of this comparison
      if ( temp.equals ( "_" ) || temp.equals ( "?" ) )
      {
	value.addElement ( new Double ( 0 ) );
      }
      else
      {
	// catch exceptions here and use NaN as the value
	try
	{
	  value.addElement ( Double.valueOf ( temp ) );
	}
	catch ( NumberFormatException ex )
	{
	  value.addElement ( new Double ( Double.NaN ) );
	}
	
      } // end else try parsing string as double
      
    } // end while more tokens
    
    // make an array to return just the double values
    double [] dArray = new double [ value.size() ];

    for ( int i = 0; i < value.size(); i++ )
    {
      dArray [ i ] = ((Double)value.elementAt ( i )).doubleValue();

    } // end for each values

    // all done
    return ( dArray );

  } // END public static double [] getValues ( String v )

//********************************************************//

  /**
   * Parses a time sequence event set attribute's value
   * into individual values (events).
   *
   * @param v
   * the string value to parse
   *
   * @return
   * an array of event values, null if v not an event set
   */

  public static String [] getEvents ( String v )
  {
    switch ( debug ) {
    case 3: case 2: case 1:
      System.out.print ( "TSUtility.getEvents ( " + v + " ) called;" );
    } // end switch debug

    // first set up a string tokenizer for sets
    StringTokenizer st = new StringTokenizer ( v, "" + SET_BEGIN +
					       SET_END + SET_DELIMITER );
    String s;

    // list of events to return
    Vector eventVector = new Vector();

    while ( st.hasMoreTokens() )
    {
      s = st.nextToken();

      // make sure this token is an event
      if ( ! isTimeSequenceEvent ( s ) )
      {
	return ( null );

      } // end if not an event

      eventVector.add ( s );

    } // end while more tokens

    String [] eventArray = new String [ eventVector.size() ];

    for ( int i = 0; i < eventArray.length; i++ )
    {
      eventArray [ i ] = (String)eventVector.elementAt ( i );

    } // end for each event

    // all done
    return ( eventArray );

  } // END public static String [] getEvents ( String v )

//********************************************************//

  /**
   * Tests if the value is a time sequence.
   * Any value with set delimiters will NOT be interpreted
   * as a time sequence.
   *
   * @param v
   * the value to test
   *
   * @return
   * true if the value is a time sequence, false otherwise
   */

  public static boolean isTimeSequence ( String v )
  {
    // does the delimiter appear?
    if ( v.indexOf ( TIME_SEQUENCE_DELIMITER ) > -1 )
    {
      // make sure a set delimiter does NOT appear
      if ( ( v.indexOf ( SET_BEGIN ) < 0 ) &&
	   ( v.indexOf ( SET_END ) < 0 ) &&
	   ( v.indexOf ( SET_DELIMITER ) < 0 ) )
      {
	return ( true );
	
      } // end if a set delimiter does NOT appear
      
    } // end if we found the delimiter
    
    // I guess it isn't a time sequence
    return ( false );

  } // END public static boolean isTimeSequence ( String v )
  
//********************************************************//

  /**
   * Tests if the value is a time sequence event.
   * The value of a time sequence event is a time sequence
   * with only two values.
   *
   * @param v
   * the value to test
   *
   * @return
   * true if the value is a time sequence event, false
   * otherwise
   */

  public static boolean isTimeSequenceEvent ( String v )
  {
    // make sure a set delimeter does NOT appear
    if ( ( v.indexOf ( SET_BEGIN ) < 0 ) &&
	 ( v.indexOf ( SET_END ) < 0 ) &&
	 ( v.indexOf ( SET_DELIMITER ) < 0 ) )
    {
      // does the time sequence delimiter appear?
      int firstD = v.indexOf ( TIME_SEQUENCE_DELIMITER );

      if ( firstD > -1 )
      {
	// make sure it appears no more
	if ( v.indexOf ( TIME_SEQUENCE_DELIMITER, firstD + 1 ) < 0 )
	{
	  return ( true );
	} // end if there is only one time sequence delimiter

      } // end if we found the delimiter
	
    } // end if a set delimiter does NOT appear
    
    // I guess it isn't a time sequence event
    return ( false );

  } // END public static boolean isTimeSequenceEvent ( String v )
  
//********************************************************//

  /**
   * Tests if the value is a set of time sequence events.
   * The value of a time sequence event is a time sequence
   * with only two values. A set uses a set delimiter to
   * seperate the events in a string. Note: A set can have
   * a single element.
   *
   * @param v
   * the value to test
   *
   * @return
   * true if the value is a set of time sequence events,
   * false otherwise
   */

  public static boolean isTimeSequenceEventSet ( String v )
  {
    // first set up a string tokenizer for sets
    StringTokenizer st = new StringTokenizer ( v, "" + SET_BEGIN +
					       SET_END + SET_DELIMITER );

    String s;

    while ( st.hasMoreTokens() )
    {
      s = st.nextToken();

      // make sure this token is an event
      if ( ! isTimeSequenceEvent ( s ) )
      {
	switch ( debug ) {
	case 3: case 2: //case 0:
	  System.out.println ( "TSUtility.isTimeSequenceEventSet: " +
			       v + " is not an event set because: " +
			       s + " is not an event. (" +
			       isTimeSequenceEvent ( s ) + ")" );
	} // end switch debug

	return ( false );

      } // end if not an event

    } // end while more tokens

    // I guess it is a time sequence event set
    return ( true );

  } // END public static boolean isTimeSequenceEvent ( String v )
  
//********************************************************//
  
  /**
   * Main method for testing this class.
   *
   * @param argv
   */

  public static void main ( String [] argv ) 
  {
    System.out.println ( "isTimeSequence ( 23:24:4566:234 ) : " +
			 isTimeSequence ( "23:24:4566:234" ) );

    System.out.println ( "isTimeSequence ( bam baby ) : " +
			 isTimeSequence ( "bam baby" ) );

    System.out.println ( "isTimeSequence ( bam baby:boo baby ) : " +
			 isTimeSequence ( "bam baby:boo baby" ) );

    System.out.println ( "isTimeSequence ( {bam baby^boo baby} ) : " +
			 isTimeSequence ( "{bam baby^boo baby}" ) );

    System.out.println ( "isTimeSequenceEvent ( 23:24:4566:234 ) : " +
			 isTimeSequenceEvent ( "23:24:4566:234" ) );

    System.out.println ( "isTimeSequenceEvent ( 23:234 ) : " +
			 isTimeSequenceEvent ( "23:234" ) );

  } // END public static void main ( String [] argv )

//********************************************************//

} // END public class TSUtility
